import binascii
import io
from typing import Optional

from structlog import get_logger

from ...extractors import Command
from ...file_utils import Endian, convert_int32
from ...models import (
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

logger = get_logger()

# CPP/7zip/Archive/ArjHandler.cpp IsArc_Arj()
MIN_BLOCK_SIZE = 30
MAX_BLOCK_SIZE = 2600
BASIC_HEADER_SIZE = 4


class ARJError(Exception):
    pass


class InvalidARJSize(ARJError):
    """Invalid size fields in ARJ header."""


class ARJChecksumError(ARJError):
    """Main ARJ header checksum missmatch."""


class ARJExtendedHeader(ARJError):
    """Main ARJ header contains extended_header, which we don't handle."""


class ARJHandler(StructHandler):
    NAME = "arj"

    PATTERNS = [HexString("60 EA [5] 0? [2] 0?")]

    # https://docs.fileformat.com/compression/arj/
    # https://github.com/tripsin/unarj/blob/master/UNARJ.H#L203
    C_DEFINITIONS = r"""
        typedef struct basic_header {
            uint16 id;
            uint16 size;
        } basic_header_t;

        typedef struct arj_header
        {
            basic_header_t header;
            uint8 first_hdr_size; // size up to "extra data"
            uint8 archive_version;
            uint8 min_version;
            uint8 host_os; // 0-9
            uint8 arj_flags; // 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40
            uint8 security_version; // "2 = current"
            uint8 file_type; // 0-4
            uint8 garble_password;
            uint32 datetime_created;
            uint32 datetime_modified;
            uint32 archive_size;
            uint32 filepos_security_env_data;
            uint16 reserved1;
            uint16 reserved2;
            uint16 security_env_length;
            uint16 host_data;
        } arj_header_t;

        typedef struct file_header {
            basic_header_t header;
            uint8 first_hdr_size; // size up to "extra data"
            uint8 archive_version;
            uint8 min_version;
            uint8 host_os; // 0-9
            uint8 arj_flags; // 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40
            uint8 method; // 0-4
            uint8 file_type;
            uint8 garble_password;
            uint32 datetime_modified;
            uint32 compressed_size;
            uint32 original_size;
            uint32 original_file_crc;
            uint16 entryname_pos_in_filename;
            uint16 file_access_mode;
            uint16 host_data;
        } file_header_t;

        typedef struct metadata {
            char filename[];
            char comment[];
            uint32 crc;
        } metadata_t;

        typedef struct extended_header {
            uint16 size;
            // More would go here if there were an extended header
        } extended_header_t;
    """

    HEADER_STRUCT = "arj_header_t"

    EXTRACTOR = Command("7z", "x", "-y", "{inpath}", "-o{outdir}")

    DOC = HandlerDoc(
        name="ARJ",
        description="ARJ is a legacy compressed archive formats used to store multiple files with metadata such as file size, creation date, and CRC.",
        handler_type=HandlerType.ARCHIVE,
        vendor=None,
        references=[
            Reference(
                title="ARJ File Format Documentation",
                url="https://docs.fileformat.com/compression/arj/",
            ),
            Reference(
                title="ARJ Technical Information",
                url="https://github.com/tripsin/unarj/blob/master/UNARJ.H#L203",
            ),
        ],
        limitations=[],
    )

    def _read_arj_main_header(self, file: File, start_offset: int) -> int:
        file.seek(start_offset)
        main_header = self.cparser_le.arj_header(file)
        logger.debug("Main header parsed", header=main_header, _verbosity=3)

        if (
            main_header.header.size < MIN_BLOCK_SIZE
            or main_header.header.size > MAX_BLOCK_SIZE
            or main_header.header.size < main_header.first_hdr_size
        ):
            raise InvalidARJSize

        file.seek(start_offset + BASIC_HEADER_SIZE)
        content = file.read(main_header.header.size)
        calculated_crc = binascii.crc32(content)
        crc = convert_int32(file.read(4), endian=Endian.LITTLE)

        if crc != calculated_crc:
            raise ARJChecksumError

        file.seek(start_offset + main_header.first_hdr_size + BASIC_HEADER_SIZE)
        self._read_headers(file)
        return file.tell()

    def _read_arj_files(self, file: File) -> int:
        while True:
            start = file.tell()
            basic_header = self.cparser_le.basic_header(file)
            logger.debug("Basic header parsed", header=basic_header, _verbosity=3)

            if basic_header.size == 0:
                # We've reached the final empty file header. This is where we want to be.
                return file.tell()

            file.seek(start)
            file_header = self.cparser_le.file_header_t(file)

            file.seek(start + file_header.first_hdr_size + len(basic_header))
            self._read_headers(file)
            # Seek past the file contents
            file.seek(file_header.compressed_size, io.SEEK_CUR)

    def _read_headers(self, file):
        metadata = self.cparser_le.metadata_t(file)
        logger.debug("Metadata header parsed", header=metadata, _verbosity=3)

        # Lack of support for extended header is ok given that no versions of ARJ use the extended header.
        # Source: 'ARJ TECHNICAL INFORMATION', September 2001
        extended_header = self.cparser_le.extended_header_t(file)
        logger.debug("Extended header parsed", header=extended_header, _verbosity=3)
        if extended_header.size != 0:
            raise ARJExtendedHeader

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        try:
            # Read past the main header.
            self._read_arj_main_header(file, start_offset)
            end_of_arj = self._read_arj_files(file)
        except ARJError as exc:
            logger.debug(
                "Invalid ARJ file",
                start_offset=start_offset,
                reason=exc.__doc__,
                _verbosity=2,
            )
            return None

        return ValidChunk(
            start_offset=start_offset,
            end_offset=end_of_arj,
        )



import os
from pathlib import Path
from typing import Optional

import arpy
from structlog import get_logger

from ...file_utils import FileSystem, OffsetFile, iterate_file
from ...models import (
    Extractor,
    ExtractResult,
    File,
    Handler,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    ValidChunk,
)
from ...report import ExtractionProblem

logger = get_logger()


HEADER_LENGTH = 0x44
SIGNATURE_LENGTH = 0x8


class ArExtractor(Extractor):
    def extract(self, inpath: Path, outdir: Path) -> Optional[ExtractResult]:
        fs = FileSystem(outdir)

        with arpy.Archive(inpath.as_posix()) as archive:
            archive.read_all_headers()

            for name in sorted(archive.archived_files):
                archived_file = archive.archived_files[name]

                try:
                    path = Path(name.decode())
                except UnicodeDecodeError:
                    path = Path(name.decode(errors="replace"))
                    fs.record_problem(
                        ExtractionProblem(
                            path=repr(name),
                            problem="Path is not a valid UTF/8 string",
                            resolution=f"Converted to {path}",
                        )
                    )

                fs.write_chunks(
                    path,
                    chunks=iterate_file(
                        archived_file,
                        0,
                        archived_file.header.size,
                    ),
                )

        return ExtractResult(reports=fs.problems)


class ARHandler(Handler):
    NAME = "ar"

    PATTERNS = [
        HexString(
            """
            // "!<arch>\\n", 58 chars of whatever, then the ARFMAG
            21 3C 61 72 63 68 3E 0A [58] 60 0A
            """
        )
    ]

    EXTRACTOR = ArExtractor()

    DOC = HandlerDoc(
        name="AR",
        description="Unix AR (archive) files are used to store multiple files in a single archive with a simple header format.",
        handler_type=HandlerType.ARCHIVE,
        vendor=None,
        references=[
            Reference(
                title="Unix AR File Format Documentation",
                url="https://en.wikipedia.org/wiki/Ar_(Unix)",
            )
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        offset_file = OffsetFile(file, start_offset)
        ar = arpy.Archive(fileobj=offset_file)  # type: ignore

        try:
            ar.read_all_headers()
        except arpy.ArchiveFormatError as exc:
            logger.debug(
                "Hit an ArchiveFormatError, we've probably hit some other kind of data",
                exc_info=exc,
            )

            # wind the cursor back the whole header length to check if we failed on
            # the first match, which means malformed AR archive
            ar.file.seek(-HEADER_LENGTH, os.SEEK_CUR)
            # we check if we failed on the first match
            if start_offset == file.tell():
                return None
            # otherwise we seek past the signature (failure on malformed AR archive
            # within the whole file, not at the start)
            ar.file.seek(SIGNATURE_LENGTH, os.SEEK_CUR)

        return ValidChunk(
            start_offset=start_offset,
            end_offset=file.tell(),
        )



from typing import Optional

from pyfatfs._exceptions import PyFATException
from pyfatfs.PyFat import PyFat
from structlog import get_logger

from unblob.extractors.command import Command
from unblob.file_utils import InvalidInputFormat

from ...models import (
    File,
    Handler,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    ValidChunk,
)

logger = get_logger()


class PyFatNoClose(PyFat):
    def close(self):
        return


def get_max_offset(fs: PyFat) -> int:
    # see PyFat.get_cluster_chain for context
    i = len(fs.fat)
    while i > 0 and fs.fat[i - 1] == 0:
        i -= 1
    # i - 1 is the last cluster that is part of a filesystem object
    return fs.get_data_cluster_address(i)


class FATHandler(Handler):
    NAME = "fat"

    PATTERNS = [
        HexString(
            """
            // An initial x86 short jump instruction
            // OEMName (8 bytes)
            // BytesPerSec (2 bytes)
            // SecPerClus (1 byte) "Must be one of 1, 2, 4, 8, 16, 32, 64, 128."
            // 495 (0x1EF) bytes of whatever
            // 55 AA is the "signature". "This will be the end of the sector only in case the
            // sector size is 512."
            ( EB | E9 ) [13] ( 01 | 02 | 04 | 08 | 10 | 20 | 40 | 80 ) [495] 55 AA
        """
        )
    ]

    EXTRACTOR = Command("7z", "x", "-y", "{inpath}", "-o{outdir}")

    DOC = HandlerDoc(
        name="FAT",
        description="FAT (File Allocation Table) is a file system format used for organizing and managing files on storage devices, supporting FAT12, FAT16, and FAT32 variants. It uses a table to map file clusters, enabling efficient file storage and retrieval.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="FAT Wikipedia",
                url="https://en.wikipedia.org/wiki/File_Allocation_Table",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        pyfat_fs = PyFatNoClose(offset=start_offset)
        try:
            pyfat_fs.set_fp(file)  # type: ignore
        except PyFATException as e:
            raise InvalidInputFormat from e

        # we have exactly one of these set to non-0, depending on FAT version
        total_sectors = max(
            pyfat_fs.bpb_header["BPB_TotSec16"],
            pyfat_fs.bpb_header["BPB_TotSec32"],
        )

        size = total_sectors * pyfat_fs.bpb_header["BPB_BytsPerSec"]
        file_size = file.size()
        if start_offset + size > file_size:
            size = get_max_offset(pyfat_fs)

        return ValidChunk(
            start_offset=start_offset,
            end_offset=start_offset + size,
        )



from pathlib import Path
from typing import Optional

from structlog import get_logger

from unblob.file_utils import Endian, File, InvalidInputFormat, StructParser
from unblob.models import (
    Extractor,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

logger = get_logger()

C_DEFINITIONS = r"""
    typedef struct engenius_header {
        uint32 unknown_1;
        uint32 vendor_id;
        uint32 product_id;
        char version[20];
        uint32 length;
        uint32 unknown_2;
        char checksum[16];
        char padding[32];
        uint32 unknown_3;
        char magic[4];
        char reg_dom[8];
        uint32 major_version;
        uint32 minor_version;
        uint32 micro_version;
        uint32 release_date;
        uint32 c_major_version;
        uint32 c_minor_version;
        uint32 c_micro_version;
        uint32 model_len;
        char model[model_len];
    } engenius_header_t;
"""

XOR_KEY = b"\xac\x78\x3c\x9e\xcf\x67\xb3\x59"
XOR_KEY_LEN = len(XOR_KEY)


def decrypter(reference):
    def decrypt(value, offset):
        nonlocal reference
        return value ^ XOR_KEY[(offset - reference) % XOR_KEY_LEN]

    return decrypt


class EngeniusExtractor(Extractor):
    def __init__(self):
        self._struct_parser = StructParser(C_DEFINITIONS)

    def extract(self, inpath: Path, outdir: Path):
        outpath = outdir.joinpath(f"{inpath.name}.decrypted")

        with File.from_path(inpath) as f:
            engenius_header = self._struct_parser.parse(
                "engenius_header_t", f, Endian.BIG
            )
            logger.debug(
                "engenius_header_t",
                engenius_header=engenius_header,
                size=len(engenius_header),
                _verbosity=3,
            )
            decrypt = decrypter(f.find(XOR_KEY))
            with outpath.open("wb") as outfile:
                decrypted = bytearray()
                for offset in range(f.tell(), engenius_header.length):
                    decrypted.append(decrypt(f[offset], offset))
                outfile.write(decrypted)


class EngeniusHandler(StructHandler):
    NAME = "engenius"

    PATTERNS = [HexString("12 34 56 78 61 6c 6c")]

    C_DEFINITIONS = C_DEFINITIONS
    HEADER_STRUCT = "engenius_header_t"
    EXTRACTOR = EngeniusExtractor()
    PATTERN_MATCH_OFFSET = -0x5C

    DOC = HandlerDoc(
        name="Engenius",
        description="Engenius firmware files contain a custom header with metadata, followed by encrypted data using an XOR cipher.",
        handler_type=HandlerType.ARCHIVE,
        vendor="Engenius",
        references=[
            Reference(
                title="enfringement - Tools for working with EnGenius WiFi hardware.",
                url="https://github.com/ryancdotorg/enfringement",  # Replace with actual reference if available
            )
        ],
        limitations=["Does not support all firmware versions."],
    )

    def is_valid_header(self, header) -> bool:
        if header.length <= len(header):
            return False
        try:
            header.model.decode("utf-8")
        except UnicodeDecodeError:
            return False
        return True

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        header = self.parse_header(file, endian=Endian.BIG)

        if not self.is_valid_header(header):
            raise InvalidInputFormat("Invalid Engenius header.")

        return ValidChunk(
            start_offset=start_offset,
            end_offset=start_offset + len(header) + header.length,
        )



import ctypes
import sys
import threading
from collections.abc import Iterable
from pathlib import Path
from typing import Callable, Optional, TypeVar

from structlog import get_logger

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

from unblob._rust.sandbox import (
    AccessFS as AccessFS,
)
from unblob._rust.sandbox import (
    SandboxError as SandboxError,
)
from unblob._rust.sandbox import (
    restrict_access as restrict_access,
)
from unblob.processing import ExtractionConfig

logger = get_logger()

P = ParamSpec("P")
R = TypeVar("R")


class Sandbox:
    """Configures restricted file-systems to run functions in.

    When calling ``run()``, a separate thread will be configured with
    minimum required file-system permissions. All subprocesses spawned
    from that thread will honor the restrictions.
    """

    def __init__(
        self,
        config: ExtractionConfig,
        log_path: Path,
        report_file: Optional[Path],
        extra_passthrough: Iterable[AccessFS] = (),
    ):
        self.passthrough = [
            # Python, shared libraries, extractor binaries and so on
            AccessFS.read("/"),
            # Multiprocessing
            AccessFS.read_write("/dev/shm"),  # noqa: S108
            # Extracted contents
            AccessFS.read_write(config.extract_root),
            AccessFS.make_dir(config.extract_root.parent),
            AccessFS.read_write(log_path),
            *extra_passthrough,
        ]

        if report_file:
            self.passthrough += [
                AccessFS.read_write(report_file.parent),
            ]

    def run(self, callback: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
        """Run callback with restricted filesystem access."""
        exception = None
        result = None

        def _run_in_thread(callback, *args, **kwargs):
            nonlocal exception, result

            self._try_enter_sandbox()
            try:
                result = callback(*args, **kwargs)
            except BaseException as e:
                exception = e

        thread = threading.Thread(
            target=_run_in_thread, args=(callback, *args), kwargs=kwargs
        )
        thread.start()

        try:
            thread.join()
        except KeyboardInterrupt:
            raise_in_thread(thread, KeyboardInterrupt)
            thread.join()

        if exception:
            raise exception  # pyright: ignore[reportGeneralTypeIssues]
        return result  # pyright: ignore[reportReturnType]

    def _try_enter_sandbox(self):
        try:
            restrict_access(*self.passthrough)
        except SandboxError:
            logger.warning(
                "Sandboxing FS access is unavailable on this system, skipping."
            )


def raise_in_thread(thread: threading.Thread, exctype: type) -> None:
    if thread.ident is None:
        raise RuntimeError("Thread is not started")

    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_ulong(thread.ident), ctypes.py_object(exctype)
    )

    # success
    if res == 1:
        return

    # Need to revert the call to restore interpreter state
    ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_ulong(thread.ident), None)

    # Thread could have exited since
    if res == 0:
        return

    # Something bad have happened
    raise RuntimeError("Could not raise exception in thread", thread.ident)



from ..models import DirectoryHandlers, Handlers
from .archive import (
    ar,
    arc,
    arj,
    cab,
    cpio,
    dmg,
    partclone,
    rar,
    sevenzip,
    stuffit,
    tar,
)
from .archive import (
    zip as ziparchive,
)
from .archive.autel import ecc
from .archive.dlink import encrpted_img, shrs
from .archive.engeniustech import engenius
from .archive.hp import bdl, ipkg
from .archive.instar import bneg, instar_hd
from .archive.netgear import chk, trx
from .archive.qnap import qnap_nas
from .archive.xiaomi import hdr
from .compression import (
    bzip2,
    compress,
    gzip,
    lz4,
    lzh,
    lzip,
    lzma,
    lzo,
    uzip,
    xz,
    zlib,
    zstd,
)
from .executable import elf
from .filesystem import (
    cramfs,
    extfs,
    fat,
    iso9660,
    jffs2,
    ntfs,
    romfs,
    squashfs,
    ubi,
    yaffs,
)
from .filesystem.android import erofs, sparse

BUILTIN_HANDLERS: Handlers = (
    cramfs.CramFSHandler,
    extfs.EXTHandler,
    fat.FATHandler,
    jffs2.JFFS2NewHandler,
    jffs2.JFFS2OldHandler,
    ntfs.NTFSHandler,
    romfs.RomFSFSHandler,
    squashfs.SquashFSv1Handler,
    squashfs.SquashFSv2Handler,
    squashfs.SquashFSv3Handler,
    squashfs.SquashFSv3DDWRTHandler,
    squashfs.SquashFSv3BroadcomHandler,
    squashfs.SquashFSv3NSHandler,
    squashfs.SquashFSv4LEHandler,
    squashfs.SquashFSv4BEHandler,
    ubi.UBIHandler,
    ubi.UBIFSHandler,
    yaffs.YAFFSHandler,
    chk.NetgearCHKHandler,
    trx.NetgearTRXv1Handler,
    trx.NetgearTRXv2Handler,
    encrpted_img.EncrptedHandler,
    shrs.SHRSHandler,
    hdr.HDR1Handler,
    hdr.HDR2Handler,
    qnap_nas.QnapHandler,
    bneg.BNEGHandler,
    bdl.HPBDLHandler,
    instar_hd.InstarHDHandler,
    ipkg.HPIPKGHandler,
    sparse.SparseHandler,
    ar.ARHandler,
    arc.ARCHandler,
    arj.ARJHandler,
    cab.CABHandler,
    tar.TarUstarHandler,
    tar.TarUnixHandler,
    cpio.PortableASCIIHandler,
    cpio.PortableASCIIWithCRCHandler,
    cpio.PortableOldASCIIHandler,
    cpio.BinaryHandler,
    sevenzip.SevenZipHandler,
    rar.RarHandler,
    ziparchive.ZIPHandler,
    dmg.DMGHandler,
    iso9660.ISO9660FSHandler,
    stuffit.StuffItSITHandler,
    stuffit.StuffIt5Handler,
    bzip2.BZip2Handler,
    compress.UnixCompressHandler,
    gzip.GZIPHandler,
    lzh.LZHHandler,
    lzip.LZipHandler,
    lzo.LZOHandler,
    lzma.LZMAHandler,
    lz4.LegacyFrameHandler,
    lz4.SkippableFrameHandler,
    lz4.DefaultFrameHandler,
    xz.XZHandler,
    zstd.ZSTDHandler,
    elf.ELF32Handler,
    elf.ELF64Handler,
    zlib.ZlibHandler,
    engenius.EngeniusHandler,
    ecc.AutelECCHandler,
    uzip.UZIPHandler,
    erofs.EROFSHandler,
    partclone.PartcloneHandler,
)

BUILTIN_DIR_HANDLERS: DirectoryHandlers = (
    sevenzip.MultiVolumeSevenZipHandler,
    gzip.MultiVolumeGzipHandler,
)



import io
from typing import Optional

from structlog import get_logger

from unblob.extractors import Command

from ...file_utils import Endian, InvalidInputFormat, convert_int8
from ...models import (
    File,
    Handler,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    ValidChunk,
)

logger = get_logger()

MAGIC_LEN = 4
BLOCK_HEADER_LEN = 3
RAW_BLOCK = 0
RLE_BLOCK = 1
COMPRESSED_BLOCK = 2
DICT_ID_FIELDSIZE_MAP = [0, 1, 2, 4]
FRAME_CONTENT_FIELDSIZE_MAP = [0, 2, 4, 8]


class ZSTDHandler(Handler):
    NAME = "zstd"

    PATTERNS = [HexString("28 B5 2F FD")]

    EXTRACTOR = Command("zstd", "-d", "{inpath}", "-o", "{outdir}/zstd.uncompressed")

    DOC = HandlerDoc(
        name="ZSTD",
        description="Zstandard (ZSTD) is a fast lossless compression algorithm with high compression ratios, designed for modern data storage and transfer. Its file format includes a frame structure with optional dictionary support and checksums for data integrity.",
        handler_type=HandlerType.COMPRESSION,
        vendor=None,
        references=[
            Reference(
                title="Zstandard File Format Specification",
                url="https://facebook.github.io/zstd/zstd_manual.html",
            ),
            Reference(
                title="Zstandard Wikipedia",
                url="https://en.wikipedia.org/wiki/Zstandard",
            ),
        ],
        limitations=[],
    )

    def get_frame_header_size(self, frame_header_descriptor: int) -> int:
        single_segment = (frame_header_descriptor >> 5 & 1) & 0b1
        dictionary_id = frame_header_descriptor >> 0 & 0b11
        frame_content_size = (frame_header_descriptor >> 6) & 0b11
        return (
            int(not single_segment)
            + DICT_ID_FIELDSIZE_MAP[dictionary_id]
            + FRAME_CONTENT_FIELDSIZE_MAP[frame_content_size]
            + (single_segment and not frame_content_size)
        )

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        file.seek(start_offset, io.SEEK_SET)
        file.seek(MAGIC_LEN, io.SEEK_CUR)

        frame_header_descriptor = convert_int8(file.read(1), Endian.LITTLE)
        frame_header_size = self.get_frame_header_size(frame_header_descriptor)

        content_checksum_flag = frame_header_descriptor >> 2 & 1
        content_checksum_size = 4 if content_checksum_flag else 0

        unused_bit = frame_header_descriptor >> 4 & 1
        reserved_bit = frame_header_descriptor >> 3 & 1

        # these values MUST be zero per the standard
        if unused_bit != 0x00 or reserved_bit != 0x0:
            raise InvalidInputFormat("Invalid frame header format.")

        file.seek(frame_header_size, io.SEEK_CUR)

        last_block = False
        while not last_block:
            block_header_val = file.read(BLOCK_HEADER_LEN)
            # EOF
            if not block_header_val:
                raise InvalidInputFormat("Premature end of ZSTD stream.")
            block_header = int.from_bytes(block_header_val, byteorder="little")
            last_block = block_header >> 0 & 0b1
            block_type = block_header >> 1 & 0b11

            if block_type in [RAW_BLOCK, COMPRESSED_BLOCK]:
                block_size = block_header >> 3
            elif block_type == RLE_BLOCK:
                block_size = 1
            else:
                raise InvalidInputFormat("Invalid block type")
            file.seek(block_size, io.SEEK_CUR)

        file.seek(content_checksum_size, io.SEEK_CUR)

        return ValidChunk(
            start_offset=start_offset,
            end_offset=file.tell(),
        )



from typing import Optional

from structlog import get_logger

from ...extractors import Command
from ...models import (
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

logger = get_logger()


class CABHandler(StructHandler):
    NAME = "cab"

    PATTERNS = [HexString("4D 53 43 46 00 00 00 00 // MSCF then reserved dword")]
    C_DEFINITIONS = r"""
        typedef struct cab_header
        {
            char  signature[4];  /* cabinet file signature contains the characters 'M','S','C','F' (bytes 0x4D, 0x53, 0x43, 0x46). */
                                /* This field is used to assure that the file is a cabinet file. */
            uint32  reserved1;     /* reserved */
            uint32  cbCabinet;     /* size of this cabinet file in bytes */
            uint32  reserved2;     /* reserved */
            uint32  coffFiles;     /* offset of the first CFFILE entry */
            uint32  reserved3;     /* reserved */
            uint8  versionMinor;  /* cabinet file format version, minor */
            uint8  versionMajor;  /* cabinet file format version, major */
            uint16  cFolders;      /* number of CFFOLDER entries in this cabinet */
            uint16  cFiles;        /* number of CFFILE entries in this cabinet */
            uint16  flags;         /* cabinet file option indicators */
            uint16  setID;         /* must be the same for all cabinets in a set*/
            uint16  iCabinet;     /* number of this cabinet file in a set */
            uint16  cbCFHeader;   /* (optional) size of per-cabinet reserved area */
            uint8  cbCFFolder;   /* (optional) size of per-folder reserved area */
            uint8  cbCFData;         /* (optional) size of per-datablock reserved area */
            uint8  szCabinetPrev[];  /* (optional) name of previous cabinet file */
            uint8  szDiskPrev[];     /* (optional) name of previous disk */
            uint8  szCabinetNext[];  /* (optional) name of next cabinet file */
            uint8  szDiskNext[];     /* (optional) name of next disk */
        } cab_header_t;
    """
    HEADER_STRUCT = "cab_header_t"

    EXTRACTOR = Command("7z", "x", "-y", "{inpath}", "-o{outdir}")

    DOC = HandlerDoc(
        name="CAB",
        description="Microsoft Cabinet (CAB) archive files are used for compressed file storage and software installation.",
        handler_type=HandlerType.ARCHIVE,
        vendor="Microsoft",
        references=[
            Reference(
                title="Microsoft Cabinet File Format Documentation",
                url="https://en.wikipedia.org/wiki/Cabinet_(file_format)",
            ),
            Reference(
                title="Ubuntu Manual - cabextract",
                url="https://manpages.ubuntu.com/manpages/focal/man1/cabextract.1.html",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        header = self.parse_header(file)

        if header.cbCabinet < len(header):
            return None

        return ValidChunk(
            start_offset=start_offset,
            end_offset=start_offset + header.cbCabinet,
        )



from pathlib import Path
from typing import Optional

from structlog import get_logger

from unblob.extractors import Command

from ...file_utils import Endian, StructParser, get_endian, round_up
from ...models import (
    Extractor,
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

logger = get_logger()

PAD_SIZES = [4_096, 1_024]


class SquashFSExtractor(Extractor):
    EXECUTABLE = "sasquatch"

    def __init__(self, big_endian_magic: int):
        self.big_endian_magic = big_endian_magic

    def extract(self, inpath: Path, outdir: Path):
        with File.from_path(inpath) as file:
            endian = get_endian(file, self.big_endian_magic)

        commands_args = []

        if endian == Endian.BIG:
            commands_args.append("-be")
        else:
            commands_args.append("-le")

        commands_args.extend(
            [
                "-no-exit-code",
                "-f",
                "-d",
                "{outdir}",
                "{inpath}",
            ]
        )
        extractor = Command(self.EXECUTABLE, *commands_args)
        extractor.extract(inpath, outdir)

    def get_dependencies(self) -> list[str]:
        return [self.EXECUTABLE]


class _SquashFSBase(StructHandler):
    BIG_ENDIAN_MAGIC = 0x73_71_73_68

    EXTRACTOR = SquashFSExtractor(0x73_71_73_68)

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        file.seek(start_offset)
        endian = get_endian(file, self.BIG_ENDIAN_MAGIC)
        header = self.parse_header(file, endian)

        end_of_data_offset = (start_offset + header.bytes_used) & 0xFF_FF_FF_FF
        file.seek(end_of_data_offset)
        padding = file.read(
            round_up(header.bytes_used, max(PAD_SIZES)) - header.bytes_used
        )

        for pad_size in sorted(PAD_SIZES, reverse=True):
            size = round_up(header.bytes_used, pad_size)
            padding_length = size - header.bytes_used

            if padding.startswith(b"\00" * padding_length):
                end_offset = start_offset + size
                return ValidChunk(start_offset=start_offset, end_offset=end_offset)

        return ValidChunk(
            start_offset=start_offset, end_offset=start_offset + header.bytes_used
        )


class SquashFSv1Handler(_SquashFSBase):
    NAME = "squashfs_v1"

    PATTERNS = [
        HexString(
            """
            // 00000000  73 71 73 68 00 00 00 03  00 00 00 00 00 00 00 00  |sqsh............|
            // 00000010  00 00 00 00 00 00 00 00  00 00 00 00 00 01 00 00  |................|
            // squashfs_v1_magic_be
            73 71 73 68 [24] 00 01
        """
        ),
        HexString(
            """
            // 00000000  68 73 71 73 03 00 00 00  00 00 00 00 00 00 00 00  |hsqs............|
            // 00000010  00 00 00 00 00 00 00 00  00 00 00 00 01 00 00 00  |................|
            // squashfs_v1_magic_le
            68 73 71 73 [24] 01 00
        """
        ),
    ]

    C_DEFINITIONS = r"""
        typedef struct squashfs_super_block
        {
            char   s_magic[4];
            uint32 inodes;
            uint32 bytes_used;
            uint32 uid_start;
            uint32 guid_start;
            uint32 inode_table_start;
            uint32 directory_table_start;
            uint16 s_major;
            uint16 s_minor;
        } squashfs_super_block_t;
    """
    HEADER_STRUCT = "squashfs_super_block_t"

    DOC = HandlerDoc(
        name="SquashFS (v1)",
        description="SquashFS version 1 is a compressed, read-only file system format designed for minimal storage usage. It is commonly used in embedded systems and early Linux distributions.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="SquashFS Documentation",
                url="https://dr-emann.github.io/squashfs/",
            ),
            Reference(
                title="SquashFS Wikipedia",
                url="https://en.wikipedia.org/wiki/SquashFS",
            ),
        ],
        limitations=[],
    )


class SquashFSv2Handler(SquashFSv1Handler):
    NAME = "squashfs_v2"

    PATTERNS = [
        HexString(
            """
            // 00000000  73 71 73 68 00 00 00 03  00 00 00 00 00 00 00 00  |sqsh............|
            // 00000010  00 00 00 00 00 00 00 00  00 00 00 00 00 02 00 00  |................|
            // squashfs_v2_magic_be
            73 71 73 68 [24] 00 02
        """
        ),
        HexString(
            """
            // 00000000  68 73 71 73 03 00 00 00  00 00 00 00 00 00 00 00  |hsqs............|
            // 00000010  00 00 00 00 00 00 00 00  00 00 00 00 02 00 00 00  |................|
            // squashfs_v2_magic_le
            68 73 71 73 [24] 02 00
        """
        ),
    ]

    DOC = HandlerDoc(
        name="SquashFS (v2)",
        description="SquashFS version 2 is a compressed, read-only file system format designed for minimal storage usage. It builds upon version 1 with additional features and improvements for embedded systems and Linux distributions.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="SquashFS Documentation",
                url="https://dr-emann.github.io/squashfs/",
            ),
            Reference(
                title="SquashFS Wikipedia",
                url="https://en.wikipedia.org/wiki/SquashFS",
            ),
        ],
        limitations=[],
    )


class SquashFSv3Handler(_SquashFSBase):
    NAME = "squashfs_v3"

    PATTERNS = [
        HexString(
            """
            // 00000000  73 71 73 68 00 00 00 03  00 00 00 00 00 00 00 00  |sqsh............|
            // 00000010  00 00 00 00 00 00 00 00  00 00 00 00 00 03 00 00  |................|
            // squashfs_v3_magic_be
            73 71 73 68 [24] 00 03
        """
        ),
        HexString(
            """
            // 00000000  68 73 71 73 03 00 00 00  00 00 00 00 00 00 00 00  |hsqs............|
            // 00000010  00 00 00 00 00 00 00 00  00 00 00 00 03 00 00 00  |................|
            // squashfs_v3_magic_le
            68 73 71 73 [24] 03 00
        """
        ),
    ]

    DOC = HandlerDoc(
        name="SquashFS (v3)",
        description="SquashFS version 3 is a compressed, read-only file system format designed for minimal storage usage. It is widely used in embedded systems and Linux distributions for efficient storage and fast access.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="SquashFS Documentation",
                url="https://dr-emann.github.io/squashfs/",
            ),
            Reference(
                title="SquashFS Wikipedia",
                url="https://en.wikipedia.org/wiki/SquashFS",
            ),
        ],
        limitations=[],
    )

    C_DEFINITIONS = r"""
        typedef struct squashfs3_super_block
        {
            char   s_magic[4];
            uint32 inodes;
            uint32 bytes_used_2;
            uint32 uid_start_2;
            uint32 guid_start_2;
            uint32 inode_table_start_2;
            uint32 directory_table_start_2;
            uint16 s_major;
            uint16 s_minor;
            uint16 block_size_1;
            uint16 block_log;
            uint8  flags;
            uint8  no_uids;
            uint8  no_guids;
            uint32 mkfs_time /* time of filesystem creation */;
            uint64 root_inode;
            uint32 block_size;
            uint32 fragments;
            uint32 fragment_table_start_2;
            uint64  bytes_used;
            uint64  uid_start;
            uint64  guid_start;
            uint64  inode_table_start;
            uint64  directory_table_start;
            uint64  fragment_table_start;
            uint64  lookup_table_start;
        } squashfs3_super_block_t;
    """
    HEADER_STRUCT = "squashfs3_super_block_t"


class SquashFSv3DDWRTHandler(SquashFSv3Handler):
    NAME = "squashfs_v3_ddwrt"

    BIG_ENDIAN_MAGIC = 0x74_71_73_68

    EXTRACTOR = SquashFSExtractor(0x74_71_73_68)

    PATTERNS = [
        HexString(
            """
            // 00000000  68 73 71 74 21 02 00 00  00 00 00 00 00 00 00 00  |hsqt!...........|
            // 00000010  00 00 00 00 00 00 00 00  50 02 00 00 03 00 00 00  |........P.......|
            // squashfs_v3_magic_ddwrt_le
            68 73 71 74 [24] 03 00
        """
        ),
        HexString(
            """
            // 00000000  74 71 73 68 00 00 00 05  00 00 00 00 00 00 00 00  |tqsh............|
            // 00000010  00 00 00 00 00 00 00 00  00 00 00 00 00 03 00 01  |................|
            // squashfs_v3_magic_ddwrt_be
            74 71 73 68 [24] 00 03
        """
        ),
    ]

    DOC = HandlerDoc(
        name="SquashFS (v3-DDWRT)",
        description="SquashFS version 3 DD-WRT is a variant of the SquashFS v3 format used in DD-WRT firmware. It features a unique magic number and may include specific optimizations for embedded systems.",
        handler_type=HandlerType.FILESYSTEM,
        vendor="DDWRT",
        references=[
            Reference(
                title="SquashFS Documentation",
                url="https://dr-emann.github.io/squashfs/",
            ),
            Reference(
                title="SquashFS Wikipedia",
                url="https://en.wikipedia.org/wiki/SquashFS",
            ),
        ],
        limitations=[],
    )


class SquashFSv3BroadcomHandler(SquashFSv3Handler):
    NAME = "squashfs_v3_broadcom"

    BIG_ENDIAN_MAGIC = 0x71_73_68_73

    EXTRACTOR = SquashFSExtractor(0x71_73_68_73)

    PATTERNS = [
        HexString(
            """
            // 00000000  73 68 73 71 0f 05 00 00  c8 a9 00 01 00 00 00 bc  |shsq............|
            // 00000010  1f 2d 00 a2 d0 2b 00 bf  79 2e 00 65 03 00 00 00  |.-...+..y..e....|
            // squashfs_v3_magic_broadcom_le
            73 68 73 71 [24] 03 00
        """
        ),
        HexString(
            """
            // 00000000  71 73 68 73 0f 05 00 00  c8 a9 00 01 00 00 00 bc  |qshs............|
            // 00000010  1f 2d 00 a2 d0 2b 00 bf  79 2e 00 65 00 03 00 00  |.-...+..y..e....|
            // squashfs_v3_magic_broadcom_be
            71 73 68 73 [24] 00 03

        """
        ),
    ]

    DOC = HandlerDoc(
        name="SquashFS (v3-Broadcom)",
        description="SquashFS version 3 Broadcom is a variant of the SquashFS v3 format used in Broadcom firmware. It features a unique magic number and may include specific optimizations for Broadcom devices.",
        handler_type=HandlerType.FILESYSTEM,
        vendor="Broadcom",
        references=[
            Reference(
                title="SquashFS Documentation",
                url="https://dr-emann.github.io/squashfs/",
            ),
            Reference(
                title="SquashFS Wikipedia",
                url="https://en.wikipedia.org/wiki/SquashFS",
            ),
        ],
        limitations=[],
    )


class SquashFSv3NSHandler(SquashFSv3Handler):
    NAME = "squashfs_v3_nonstandard"

    BIG_ENDIAN_MAGIC = 0x73_71_6C_7A

    EXTRACTOR = SquashFSExtractor(0x73_71_6C_7A)

    PATTERNS = [
        HexString(
            """
            // 00000000  7a 6c 71 73 00 00 04 df  57 00 17 95 46 00 19 ed  |zlqs....W...F...|
            // 00000010  20 08 04 8e 02 40 01 06  c9 02 16 00 00 03 00 01  | ....@..........|
            // squashfs_v3_magic_nonstandard_le
            7A 6c 71 73 [24] 03 00
        """
        ),
        HexString(
            """
            // 00000000  73 71 6c 7a 00 00 04 df  57 00 17 95 46 00 19 ed  |sqlz....W...F...|
            // 00000010  20 08 04 8e 02 40 01 06  c9 02 16 00 00 03 00 01  | ....@..........|
            // squashfs_v3_magic_nonstandard_be
            73 71 6c 7A [24] 00 03
        """
        ),
    ]

    DOC = HandlerDoc(
        name="SquashFS (v3-non-standard)",
        description="SquashFS version 3 is a compressed, read-only file system format designed for minimal storage usage. It is widely used in embedded systems and Linux distributions for efficient storage and fast access.",
        handler_type=HandlerType.FILESYSTEM,
        vendor="unknown",
        references=[
            Reference(
                title="SquashFS Documentation",
                url="https://dr-emann.github.io/squashfs/",
            ),
            Reference(
                title="SquashFS Wikipedia",
                url="https://en.wikipedia.org/wiki/SquashFS",
            ),
        ],
        limitations=[],
    )


class SquashFSv4LEHandler(_SquashFSBase):
    NAME = "squashfs_v4_le"

    PATTERNS = [
        HexString(
            """
            // 00000000  68 73 71 73 03 00 00 00  00 c1 9c 61 00 00 02 00  |hsqs.......a....|
            // 00000010  01 00 00 00 01 00 11 00  c0 00 01 00 04 00 00 00  |................|
            // squashfs_v4_magic_le
            68 73 71 73 [24] 04 00
        """
        ),
    ]

    C_DEFINITIONS = r"""
        typedef struct squashfs4_super_block
        {
            char   s_magic[4];
            uint32 inodes;
            uint32 mkfs_time /* time of filesystem creation */;
            uint32 block_size;
            uint32 fragments;
            uint16 compression;
            uint16 block_log;
            uint16  flags;
            uint16  no_ids;
            uint16 s_major;
            uint16 s_minor;
            uint64 root_inode;
            uint64  bytes_used;
            uint64  id_table_start;
            uint64  xattr_id_table_start;
            uint64  inode_table_start;
            uint64  directory_table_start;
            uint64  fragment_table_start;
            uint64  lookup_table_start;
        } squashfs4_super_block_t;
    """
    HEADER_STRUCT = "squashfs4_super_block_t"

    DOC = HandlerDoc(
        name="SquashFS (v4-LE)",
        description="SquashFS version 4 is a compressed, read-only file system format designed for minimal storage usage and fast access. It is widely used in embedded systems and Linux distributions for efficient storage management.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="SquashFS Documentation",
                url="https://dr-emann.github.io/squashfs/",
            ),
            Reference(
                title="SquashFS Wikipedia",
                url="https://en.wikipedia.org/wiki/SquashFS",
            ),
        ],
        limitations=[],
    )


class SquashFSv4BEExtractor(Extractor):
    EXECUTABLE = "sasquatch-v4be"

    def is_avm(self, header) -> bool:
        # see https://raw.githubusercontent.com/Freetz/freetz/master/tools/make/squashfs4-host-be/AVM-BE-format.txt
        return header.bytes_used == header.mkfs_time

    def extract(self, inpath: Path, outdir: Path):
        struct_parser = StructParser(SquashFSv4LEHandler.C_DEFINITIONS)

        with File.from_path(inpath) as f:
            header = struct_parser.parse(
                SquashFSv4LEHandler.HEADER_STRUCT, f, Endian.BIG
            )

        commands_args = []

        if not self.is_avm(header):
            commands_args.append("-be")

        commands_args.extend(
            [
                "-no-exit-code",
                "-f",
                "-d",
                "{outdir}",
                "{inpath}",
            ]
        )
        extractor = Command(self.EXECUTABLE, *commands_args)
        extractor.extract(inpath, outdir)

    def get_dependencies(self) -> list[str]:
        return [self.EXECUTABLE]


class SquashFSv4BEHandler(SquashFSv4LEHandler):
    NAME = "squashfs_v4_be"

    PATTERNS = [
        HexString(
            """
            // 00000000  73 71 73 68 03 00 00 00  00 c1 9c 61 00 00 02 00  |sqsh.......a....|
            // 00000010  01 00 00 00 01 00 11 00  c0 00 01 00 04 00 00 00  |................|
            // squashfs_v4_magic_be
            73 71 73 68 [24] 00 04
        """
        )
    ]

    EXTRACTOR = SquashFSv4BEExtractor()

    DOC = HandlerDoc(
        name="SquashFS (v4-BE)",
        description="SquashFS version 4 is a compressed, read-only file system format designed for minimal storage usage and fast access. It supports both big-endian and little-endian formats and is widely used in embedded systems and Linux distributions.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="SquashFS Documentation",
                url="https://dr-emann.github.io/squashfs/",
            ),
            Reference(
                title="SquashFS Wikipedia",
                url="https://en.wikipedia.org/wiki/SquashFS",
            ),
        ],
        limitations=[],
    )



"""LZ4 handler.

Frame format definition: https://github.com/lz4/lz4/blob/dev/doc/lz4_Frame_format.md.
"""

import io
from typing import Optional

from lz4.block import LZ4BlockError, decompress
from structlog import get_logger

from unblob.extractors import Command

from ...file_utils import Endian, InvalidInputFormat, convert_int8, convert_int32
from ...models import (
    File,
    Handler,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    ValidChunk,
)

logger = get_logger()

SKIPPABLE_FRAMES_MAGIC = [0x184D2A50 + i for i in range(16)]
FRAME_MAGIC = 0x184D2204
LEGACY_FRAME_MAGIC = 0x184C2102
FRAME_MAGICS = [*SKIPPABLE_FRAMES_MAGIC, FRAME_MAGIC, LEGACY_FRAME_MAGIC]

_1BIT = 0x01
_2BITS = 0x03

END_MARK = 0x00000000

CONTENT_SIZE_LEN = 8
BLOCK_SIZE_LEN = FRAME_SIZE_LEN = BLOCK_CHECKSUM_LEN = CONTENT_CHECKSUM_LEN = (
    MAGIC_LEN
) = DICTID_LEN = 4
FLG_LEN = BD_LEN = HC_LEN = 1
MAX_LEGACY_BLOCK_SIZE = 8 * 1024 * 1024  # 8 MB


class FLG:
    """Represents the FLG field."""

    version: int = 0
    block_independence: int = 0
    block_checksum: int = 0
    content_size: int = 0
    content_checksum: int = 0
    dictid: int = 0

    def __init__(self, raw_flg: int):
        self.version = (raw_flg >> 6) & _2BITS
        self.block_independence = (raw_flg >> 5) & _1BIT
        self.block_checksum = (raw_flg >> 4) & _1BIT
        self.content_size = (raw_flg >> 3) & _1BIT
        self.content_checksum = (raw_flg >> 2) & _1BIT
        self.dictid = raw_flg & _1BIT

    def as_dict(self) -> dict:
        return {
            "version": self.version,
            "block_independence": self.block_independence,
            "block_checksum": self.block_checksum,
            "content_size": self.content_size,
            "content_checksum": self.content_checksum,
            "dictid": self.dictid,
        }


class _LZ4HandlerBase(Handler):
    """A common base for all LZ4 formats."""

    def _skip_magic_bytes(self, file: File):
        file.seek(MAGIC_LEN, io.SEEK_CUR)

    EXTRACTOR = Command("lz4", "--decompress", "{inpath}", "{outdir}/lz4.uncompressed")


class LegacyFrameHandler(_LZ4HandlerBase):
    NAME = "lz4_legacy"
    PATTERNS = [HexString("02 21 4C 18")]

    DOC = HandlerDoc(
        name="LZ4 (legacy)",
        description="LZ4 legacy format is an older framing format used prior to the LZ4 Frame specification, featuring a simpler structure and no support for skippable frames or extensive metadata. Unlike the default LZ4 Frame format, it lacks built-in checksums, versioning, or block independence flags, making it less robust and primarily used for backward compatibility.",
        handler_type=HandlerType.COMPRESSION,
        vendor=None,
        references=[
            Reference(
                title="LZ4 Frame Format Documentation",
                url="https://github.com/lz4/lz4/blob/dev/doc/lz4_Frame_format.md",
            ),
            Reference(
                title="LZ4 Wikipedia",
                url="https://en.wikipedia.org/wiki/LZ4_(compression_algorithm)",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        self._skip_magic_bytes(file)

        while True:
            # The last block is detected either because it is followed by the “EOF” (End of File) mark,
            # or because it is followed by a known Frame Magic Number.
            raw_bsize = file.read(BLOCK_SIZE_LEN)
            if raw_bsize == b"":  # EOF
                break

            block_compressed_size = convert_int32(raw_bsize, Endian.LITTLE)
            if block_compressed_size in FRAME_MAGICS:
                # next magic, read too far
                file.seek(-4, io.SEEK_CUR)
                break

            compressed_block = file.read(block_compressed_size)
            try:
                uncompressed_block = decompress(compressed_block, MAX_LEGACY_BLOCK_SIZE)
            except LZ4BlockError:
                raise InvalidInputFormat("Invalid LZ4 legacy frame.") from None

            # See 'fixed block size' in https://android.googlesource.com/platform/external/lz4/+/HEAD/doc/lz4_Frame_format.md#legacy-frame
            if len(uncompressed_block) < MAX_LEGACY_BLOCK_SIZE:
                break

        end_offset = file.tell()
        return ValidChunk(start_offset=start_offset, end_offset=end_offset)


class SkippableFrameHandler(_LZ4HandlerBase):
    """Can be anything, basically uncompressed data."""

    NAME = "lz4_skippable"
    PATTERNS = [HexString("5? 2A 4D 18")]

    DOC = HandlerDoc(
        name="LZ4 (skippable)",
        description="LZ4 skippable format is designed to encapsulate arbitrary data within an LZ4 stream allowing compliant parsers to skip over it safely. This format does not contain compressed data itself but is often used for embedding metadata or non-LZ4 content alongside standard frames.",
        handler_type=HandlerType.COMPRESSION,
        vendor=None,
        references=[
            Reference(
                title="LZ4 Frame Format Documentation",
                url="https://github.com/lz4/lz4/blob/dev/doc/lz4_Frame_format.md",
            ),
            Reference(
                title="LZ4 Wikipedia",
                url="https://en.wikipedia.org/wiki/LZ4_(compression_algorithm)",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        self._skip_magic_bytes(file)
        frame_size = convert_int32(file.read(FRAME_SIZE_LEN), Endian.LITTLE)
        file.seek(frame_size, io.SEEK_CUR)
        end_offset = file.tell()
        return ValidChunk(start_offset=start_offset, end_offset=end_offset)


class DefaultFrameHandler(_LZ4HandlerBase):
    """Modern version, most frequently used."""

    NAME = "lz4_default"

    PATTERNS = [HexString("04 22 4D 18")]

    DOC = HandlerDoc(
        name="LZ4",
        description="LZ4 is a high-speed lossless compression algorithm designed for real-time data compression with minimal memory usage.",
        handler_type=HandlerType.COMPRESSION,
        vendor=None,
        references=[
            Reference(
                title="LZ4 Frame Format Documentation",
                url="https://github.com/lz4/lz4/blob/dev/doc/lz4_Frame_format.md",
            ),
            Reference(
                title="LZ4 Wikipedia",
                url="https://en.wikipedia.org/wiki/LZ4_(compression_algorithm)",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        self._skip_magic_bytes(file)

        # 2. we parse the frame descriptor of dynamic size
        flg_bytes = file.read(FLG_LEN)
        raw_flg = convert_int8(flg_bytes, Endian.LITTLE)
        flg = FLG(raw_flg)
        logger.debug("Parsed FLG", **flg.as_dict())

        # skip BD (max blocksize), only useful for decoders that needs to allocate memory
        file.seek(BD_LEN, io.SEEK_CUR)

        if flg.content_size:
            file.seek(CONTENT_SIZE_LEN, io.SEEK_CUR)
        if flg.dictid:
            file.seek(DICTID_LEN, io.SEEK_CUR)

        header_checksum = convert_int8(file.read(HC_LEN), Endian.LITTLE)
        logger.debug("Header checksum (HC) read", header_checksum=header_checksum)

        # 3. we read block by block until we hit the endmarker
        while True:
            block_size = convert_int32(file.read(BLOCK_SIZE_LEN), Endian.LITTLE)
            logger.debug("block_size", block_size=block_size)
            if block_size == END_MARK:
                break
            file.seek(block_size, io.SEEK_CUR)
            if flg.block_checksum:
                file.seek(BLOCK_CHECKSUM_LEN, io.SEEK_CUR)

        # 4. we reached the endmark (0x00000000)

        # 5. if frame descriptor mentions CRC, we add CRC
        if flg.content_checksum:
            file.seek(CONTENT_CHECKSUM_LEN, io.SEEK_CUR)

        end_offset = file.tell()

        return ValidChunk(start_offset=start_offset, end_offset=end_offset)



import shutil
import statistics
from pathlib import Path
from typing import Optional

from structlog import get_logger

from unblob.extractors import Command

from ...file_utils import InvalidInputFormat, SeekError, get_endian, iterate_patterns
from ...iter_utils import get_intervals
from ...models import (
    File,
    Handler,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

logger = get_logger()


class UBIFSHandler(StructHandler):
    NAME = "ubifs"

    _BIG_ENDIAN_MAGIC = 0x06_10_18_31

    # TODO: At the moment, we only match on the UBIFS superblock. Do we also want to account for
    # cases where the first node isn't a UBIFS superblock? Would such a layout actually be valid?
    # It might be valid to be flagged, but not necessarily to be extracted.
    #
    # Since we are running the handlers against every single match, regardless of whether a
    # previous chunk has already been established. That means that, for example, if we find a
    # superblock, and then many other kinds of nodes, it will take forever to run caculate_chunk()
    # against all the other nodes, and we waste loads of time and resources.

    # magic (4 bytes), 16 bytes, node type (1 byte, 0x06 is superblock),
    # group type (1 byte), 2 nulls.
    PATTERNS = [
        HexString("31 18 10 06 [16] 06 ( 00 | 01 | 02 ) 00 00"),  # LE
        HexString("06 10 18 31 [16] 06 ( 00 | 01 | 02 ) 00 00"),  # BE
    ]

    C_DEFINITIONS = r"""
        typedef struct ubifs_ch {
            uint32 magic;
            uint32 crc;
            uint64 sqnum;
            uint32 len;
            uint8 node_type;
            uint8 group_type;
            uint8 padding[2];
        } ubifs_ch_t;

        typedef struct ubifs_sb_node {
            ubifs_ch_t ch;
            uint8 padding[2];
            uint8 key_hash;
            uint8 key_fmt;
            uint32 flags;
            uint32 min_io_size;
            uint32 leb_size;
            uint32 leb_cnt;
            uint32 max_leb_cnt;
            uint64 max_bud_bytes;
            uint32 log_lebs;
            uint32 lpt_lebs;
            uint32 orph_lebs;
            uint32 jhead_cnt;
            uint32 fanout;
            uint32 lsave_cnt;
            uint32 fmt_version;
            uint16 default_compr;
            uint8 padding1[2];
            uint32 rp_uid;
            uint32 rp_gid;
            uint64 rp_size;
            uint32 time_gran;
            uint8 uuid[16];
            uint32 ro_compat_version;
            uint8 hmac[64];
            uint8 hmac_wkm[64];
            uint16 hash_algo;
            uint8 hash_mst[64];
            uint8 padding2[3774];
        } ubifs_sb_node_t;
    """
    HEADER_STRUCT = "ubifs_sb_node_t"

    EXTRACTOR = Command("ubireader_extract_files", "{inpath}", "-w", "-o", "{outdir}")

    DOC = HandlerDoc(
        name="UBIFS",
        description="UBIFS (Unsorted Block Image File System) is a flash file system designed for raw flash memory, providing wear leveling, error correction, and power failure resilience. It operates on top of UBI volumes, which manage flash blocks on raw NAND or NOR flash devices.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="UBIFS Documentation",
                url="https://www.kernel.org/doc/html/latest/filesystems/ubifs.html",
            ),
            Reference(
                title="UBIFS Wikipedia",
                url="https://en.wikipedia.org/wiki/UBIFS",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        endian = get_endian(file, self._BIG_ENDIAN_MAGIC)
        sb_header = self.parse_header(file, endian)

        # At the moment we are only matching on superblock nodes, so we can get the size of the
        # chunk from the LEB size * LEB count.
        ubifs_length = sb_header.leb_size * sb_header.leb_cnt

        return ValidChunk(
            start_offset=start_offset,
            end_offset=start_offset + ubifs_length,
        )


class UBIExtractor(Command):
    def extract(self, inpath: Path, outdir: Path):
        super().extract(inpath, outdir)
        # ubireader_extract_images creates a superfluous directory named
        # after the UBI file (inpath here), so we simply move the files up
        # and delete the remaining directory.
        superfluous_dir_path = outdir.joinpath(inpath.name)
        for file_path in superfluous_dir_path.iterdir():
            shutil.move(file_path.as_posix(), outdir.as_posix())
        shutil.rmtree(superfluous_dir_path.as_posix())


class UBIHandler(Handler):
    NAME = "ubi"

    _UBI_EC_HEADER = b"UBI#"

    PATTERNS = [HexString("55 42 49 23 01  // UBI# and version 1")]

    EXTRACTOR = UBIExtractor("ubireader_extract_images", "{inpath}", "-o", "{outdir}")

    DOC = HandlerDoc(
        name="UBI",
        description="UBI (Unsorted Block Image) is a volume management system for raw flash devices, providing wear leveling and bad block management. It operates as a layer between the MTD subsystem and higher-level filesystems like UBIFS.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="UBI Documentation",
                url="https://www.kernel.org/doc/html/latest/driver-api/ubi.html",
            ),
            Reference(
                title="UBI Wikipedia",
                url="https://en.wikipedia.org/wiki/UBIFS#UBI",
            ),
        ],
        limitations=[],
    )

    def _guess_peb_size(self, file: File) -> int:
        # Since we don't know the PEB size, we need to guess it. At the moment we just find the
        # most common interval between every erase block header we find in the image. This _might_
        # cause an issue if we had a blob containing multiple UBI images, with different PEB sizes.
        all_ubi_eraseblock_offsets = list(iterate_patterns(file, self._UBI_EC_HEADER))

        offset_intervals = get_intervals(all_ubi_eraseblock_offsets)
        if not offset_intervals:
            raise InvalidInputFormat

        return statistics.mode(offset_intervals)

    def _walk_ubi(self, file: File, peb_size: int) -> int:
        """Walk from the start_offset, at PEB-sized intervals, until we don't hit an erase block."""
        while True:
            offset = file.tell()
            first_bytes = file.read(len(self._UBI_EC_HEADER))
            if first_bytes == b"" or first_bytes != self._UBI_EC_HEADER:
                break
            try:
                file.seek(offset + peb_size)
            except SeekError:
                break
        return offset

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        peb_size = self._guess_peb_size(file)

        logger.debug("Guessed UBI PEB size", size=peb_size)

        file.seek(start_offset)
        # We don't want to parse headers, because we don't know what third party tools are doing,
        # and it would be too expensive to validate the CRC and/or calculate all of the headers
        # This is good enough and way faster than parsing headers
        end_offset = self._walk_ubi(file, peb_size)

        return ValidChunk(start_offset=start_offset, end_offset=end_offset)



from typing import Optional

from structlog import get_logger

from unblob.extractors import Command
from unblob.file_utils import Endian

from ...models import (
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

logger = get_logger()


# The system area, the first 32,768 data bytes of the disc (16 sectors of 2,048 bytes each),
# is unused by ISO 9660 and therefore available for other uses.
SYSTEM_AREA_SIZE = 0x8000


def from_733(u: bytes) -> int:
    """Convert from ISO 9660 7.3.3 format to uint32_t.

    Return the little-endian part always, to handle non-specs-compliant images.
    """
    return u[0] | (u[1] << 8) | (u[2] << 16) | (u[3] << 24)


def from_723(u: bytes) -> int:
    """Convert from ISO 9660 7.2.3 format to uint16_t.

    Return the little-endian part always, to handle non-specs-compliant images.
    """
    return u[0] | (u[1] << 8)


class ISO9660FSHandler(StructHandler):
    NAME = "iso9660"

    #
    # Match on volume descriptor type, followed by ISO_STANDARD_ID, which corresponds to the beginning of a volume descriptor.
    #
    # Volume descriptor types can be:
    #     - 0x00	Boot record volume descriptor
    #     - 0x01	Primary volume descriptor
    #     - 0x02	Supplementary volume descriptor, or enhanced volume descriptor
    #     - 0x03	Volume partition descriptor
    #     - 0xFF Volume descriptor terminator

    PATTERNS = [
        HexString(
            "( 00 | 01 | 02 | 03 ) 43 44 30 30 31 // vd_type + 'CD001' (ISO_STANDARD_ID within primary volume descriptor)"
        )
    ]

    C_DEFINITIONS = r"""
        typedef struct iso9660_dtime_s {
            uint8 dt_year;
            uint8 dt_month;
            uint8 dt_day;
            uint8 dt_hour;
            uint8 dt_minute;
            uint8 dt_second;
            int8 dt_gmtoff;
        } iso9660_dtime_t;

        typedef struct iso9660_ltime_s {
            char lt_year[4];
            char lt_month[2];
            char lt_day[2];
            char lt_hour[2];
            char lt_minute[2];
            char lt_second[2];
            char lt_hsecond[2];
            int8 lt_gmtoff;
        } iso9660_ltime_t;

        typedef struct iso9660_dir_s {
            uint8 length;
            uint8 xa_length;
            uint64 extent;
            uint64 size;
            iso9660_dtime_t recording_time;
            uint8 file_flags;
            uint8 file_unit_size;
            uint8 interleave_gap;
            uint32 volume_sequence_number;
            union {
            uint8 len;
            char str[1];
            } filename;
        } iso9660_dir_t;

        typedef struct iso9660_pvd_s {
            uint8         type;                         /**< ISO_VD_PRIMARY - 1 */
            char             id[5];                        /**< ISO_STANDARD_ID "CD001"
                                                            */
            uint8         version;                      /**< value 1 for ECMA 119 */
            char             unused1[1];                   /**< unused - value 0 */
            char          system_id[32]; /**< each char is an achar */
            char          volume_id[32]; /**< each char is a dchar */
            uint8          unused2[8];                   /**< unused - value 0 */
            /**uint64         volume_space_size;            /**< total number of
                                                            sectors */
            uint8       volume_space_size[8];
            char          unused3[32];                  /**< unused - value 0 */
            uint32         volume_set_size;              /**< often 1 */
            uint32         volume_sequence_number;       /**< often 1 */
            uint8         logical_block_size[4];           /**< sector size, e.g. 2048 */
        } iso9660_pvd_t;
    """

    HEADER_STRUCT = "iso9660_pvd_t"

    EXTRACTOR = Command("7z", "x", "-y", "{inpath}", "-o{outdir}")

    DOC = HandlerDoc(
        name="ISO 9660",
        description="ISO 9660 is a file system standard for optical disc media, defining a volume descriptor structure and directory hierarchy. It is widely used for CD-ROMs and supports cross-platform compatibility.",
        handler_type=HandlerType.FILESYSTEM,
        vendor=None,
        references=[
            Reference(
                title="ISO 9660 Specification",
                url="https://wiki.osdev.org/ISO_9660",
            ),
            Reference(
                title="ISO 9660 Wikipedia",
                url="https://en.wikipedia.org/wiki/ISO_9660",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        header = self.parse_header(file, Endian.LITTLE)
        size = from_733(header.volume_space_size) * from_723(header.logical_block_size)

        # We need to subtract the system area given that we matched on volume descriptor,
        # which is the first struct afterward.
        real_start_offset = start_offset - SYSTEM_AREA_SIZE
        if real_start_offset < 0:
            logger.warning("Invalid ISO 9660 file", offset=real_start_offset, size=size)
            return None

        return ValidChunk(
            start_offset=real_start_offset,
            end_offset=real_start_offset + size,
        )



"""Stuffit handlers.

We currently support detecting two specific Stuffit file formats:

    - Stuffit SIT (with 'SIT!' magic)

    - Stuffit 5 (with 'StuffIt (c)1997-' magic)

Due to lack of access to sample files, source code, and really old
packing tools, we don't support the following Stuffit file formats at
the moment:

    - Stuffit X ('StuffIt!' or 'StuffIt?' magic)

    - Stuffit Delux ('SITD' magic)

If you have the resources to add support for these archive formats,
feel free to do so !
"""

from typing import Optional

from structlog import get_logger

from ...extractors import Command
from ...file_utils import Endian
from ...models import (
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

logger = get_logger()


class _StuffItHandlerBase(StructHandler):
    """A common base for all StuffIt formats."""

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        header = self.parse_header(file, endian=Endian.BIG)

        return ValidChunk(
            start_offset=start_offset,
            end_offset=start_offset + header.archive_length,
        )

    EXTRACTOR = Command("unar", "-no-directory", "-o", "{outdir}", "{inpath}")


class StuffItSITHandler(_StuffItHandlerBase):
    NAME = "stuffit"

    PATTERNS = [
        HexString(
            """
            // "SIT!\\x00", then 6 bytes (uint16 number of files and uint32 size), then "rLau".
            53 49 54 21 [6] 72 4C 61 75
        """
        )
    ]

    C_DEFINITIONS = r"""
        typedef struct sit_header
        {
            char signature[4];
            uint16 num_files;
            uint32 archive_length;
            char signature2[4];
        } sit_header_t;
    """
    HEADER_STRUCT = "sit_header_t"

    EXTRACTOR = Command("unar", "-no-directory", "-o", "{outdir}", "{inpath}")

    DOC = HandlerDoc(
        name="Stuffit SIT",
        description="StuffIt SIT archives is a legacy compressed archive format commonly used on macOS and earlier Apple systems.",
        handler_type=HandlerType.ARCHIVE,
        vendor="StuffIt Technologies",
        references=[
            Reference(
                title="StuffIt SIT File Format Documentation",
                url="https://en.wikipedia.org/wiki/StuffIt",
            )
        ],
        limitations=[],
    )


class StuffIt5Handler(_StuffItHandlerBase):
    NAME = "stuffit5"

    PATTERNS = [
        HexString(
            """
            // "StuffIt (c)1997-"
            53 74 75 66 66 49 74 20 28 63 29 31 39 39 37 2D
        """
        )
    ]

    C_DEFINITIONS = r"""
        typedef struct stuffit5_header
        {
            char signature[80];
            uint32 unknown;
            uint32 archive_length;
            uint32 entry_offset;
            uint16 num_root_dir_entries;
            uint32 first_entry_offset;
        } stuffit5_header_t;
    """
    HEADER_STRUCT = "stuffit5_header_t"

    EXTRACTOR = Command("unar", "-no-directory", "-o", "{outdir}", "{inpath}")

    DOC = HandlerDoc(
        name="Stuffit SIT (v5)",
        description="StuffIt SIT archives is a legacy compressed archive format commonly used on macOS and earlier Apple systems.",
        handler_type=HandlerType.ARCHIVE,
        vendor="StuffIt Technologies",
        references=[
            Reference(
                title="StuffIt SIT File Format Documentation",
                url="https://en.wikipedia.org/wiki/StuffIt",
            )
        ],
        limitations=[],
    )



import io
from pathlib import Path
from typing import Optional

import attrs
from pyperscan import Flag, Pattern, Scan, StreamDatabase
from structlog import get_logger

from unblob.file_utils import File, iterate_file, stream_scan
from unblob.models import (
    Endian,
    Extractor,
    Handler,
    HandlerDoc,
    HandlerType,
    HexString,
    StructParser,
    ValidChunk,
)

logger = get_logger()

FOOTER_LEN = 74
SECRET = "QNAPNASVERSION"  # noqa: S105

C_DEFINITIONS = """
    typedef struct qnap_header {
        char    magic[6];
        uint32  encrypted_len;
        char    device_id[16];
        char    file_version[16];
        char    firmware_date[16];
        char    revision[16];
    } qnap_header_t;
"""
FOOTER_PATTERN = [
    HexString("69 63 70 6e 61 73"),  # encrypted gzip stream start bytes
]


@attrs.define
class QTSSearchContext:
    start_offset: int
    file: File
    end_offset: int


def is_valid_header(header) -> bool:
    try:
        header.device_id.decode("utf-8")
        header.file_version.decode("utf-8")
        header.firmware_date.decode("utf-8")
        header.revision.decode("utf-8")
    except UnicodeDecodeError:
        return False
    return True


def _hyperscan_match(
    context: QTSSearchContext, pattern_id: int, offset: int, end: int
) -> Scan:
    del pattern_id, end  # unused arguments
    if offset < context.start_offset:
        return Scan.Continue
    context.file.seek(offset, io.SEEK_SET)
    struct_parser = StructParser(C_DEFINITIONS)
    header = struct_parser.parse("qnap_header_t", context.file, Endian.LITTLE)
    logger.debug("qnap_header_t", header=header)

    if is_valid_header(header):
        context.end_offset = context.file.tell()
        return Scan.Terminate
    return Scan.Continue


def build_stream_end_scan_db(pattern_list):
    return StreamDatabase(
        *(Pattern(p.as_regex(), Flag.SOM_LEFTMOST, Flag.DOTALL) for p in pattern_list)
    )


hyperscan_stream_end_magic_db = build_stream_end_scan_db(FOOTER_PATTERN)


class QnapExtractor(Extractor):
    def __init__(self):
        self._struct_parser = StructParser(C_DEFINITIONS)

    def extract(self, inpath: Path, outdir: Path):
        outpath = outdir.joinpath(f"{inpath.name}.decrypted")
        with File.from_path(inpath) as file:
            file.seek(-FOOTER_LEN, io.SEEK_END)
            header = self._struct_parser.parse("qnap_header_t", file, Endian.LITTLE)
            eof = file.tell()
            cryptor = Cryptor(SECRET + header.file_version.decode("utf-8")[0])
            with outpath.open("wb") as outfile:
                for chunk in iterate_file(file, 0, header.encrypted_len, 1024):
                    outfile.write(cryptor.decrypt_chunk(chunk))
                for chunk in iterate_file(
                    file,
                    header.encrypted_len,
                    eof - FOOTER_LEN - header.encrypted_len,
                    1024,
                ):
                    outfile.write(chunk)


class QnapHandler(Handler):
    NAME = "qnap_nas"

    PATTERNS = [
        HexString("F5 7B 47 03"),
    ]
    EXTRACTOR = QnapExtractor()

    DOC = HandlerDoc(
        name="QNAP NAS",
        description="QNAP NAS firmware files consist of a custom header, encrypted data sections, and a footer marking the end of the encrypted stream. The header contains metadata such as device ID, firmware version, and encryption details.",
        handler_type=HandlerType.ARCHIVE,
        vendor="QNAP",
        references=[],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        context = QTSSearchContext(start_offset=start_offset, file=file, end_offset=-1)

        try:
            scanner = hyperscan_stream_end_magic_db.build(context, _hyperscan_match)  # type: ignore
            stream_scan(scanner, file)
        except Exception as e:
            logger.debug(
                "Error scanning for QNAP patterns",
                error=e,
            )
        if context.end_offset > 0:
            return ValidChunk(start_offset=start_offset, end_offset=context.end_offset)
        return None


# https://gist.github.com/ulidtko/966277a465f1856109b2d2674dcee741#file-qnap-qts-fw-cryptor-py-L114
class Cryptor:
    def __init__(self, secret):
        self.secret = list(bytes(secret, "ascii"))
        self.n = len(secret) // 2
        if self.n % 2 == 0:
            self.secret.append(0)
        self.precompute_k()
        self.acc = 0
        self.y = 0
        self.z = 0

    def scan(self, f, xs, s0):
        s = s0
        for x in xs:
            w, s = f(s, x)
            yield w

    def promote(self, char):
        return char if char < 0x80 else char - 0x101

    def precompute_k(self):
        self.k = {acc: self.table_for_acc(acc) for acc in range(256)}

    def table_for_acc(self, a):
        ks = [
            0xFFFFFFFF
            & (
                (self.promote(self.secret[2 * i] ^ a) << 8)
                + (self.secret[2 * i + 1] ^ a)
            )
            for i in range(self.n)
        ]

        def kstep(st, q):
            x = st ^ q
            y = self.lcg(x)
            z = 0xFFFF & (0x15A * x)
            return (z, y), y

        return list(self.scan(kstep, ks, 0))

    def lcg(self, x):
        return 0xFFFF & (0x4E35 * x + 1)

    def kdf(self):
        """self.secret -> 8bit hash (+ state effects)."""
        tt = self.k[self.acc]
        res = 0
        for i in range(self.n):
            yy = self.y
            self.y, t2 = tt[i]
            self.z = 0xFFFF & (self.y + yy + 0x4E35 * (self.z + i))
            res = res ^ t2 ^ self.z
        hi, lo = res >> 8, res & 0xFF
        return hi ^ lo

    def decrypt_byte(self, v):
        k = self.kdf()
        r = 0xFF & (v ^ k)
        self.acc = self.acc ^ r
        return r

    def decrypt_chunk(self, chunk):
        return bytes(map(self.decrypt_byte, chunk))



from typing import Optional

from structlog import get_logger

from unblob.file_utils import InvalidInputFormat

from ...extractors import Command
from ...models import (
    File,
    HandlerDoc,
    HandlerType,
    Reference,
    Regex,
    StructHandler,
    ValidChunk,
)

logger = get_logger()


class NTFSHandler(StructHandler):
    NAME = "ntfs"

    PATTERNS = [
        # An initial x86 short jump instruction (EB 52 90)
        # OEMName (8 bytes) = 'NTFS    '
        # bytes per sector, the rule enforce a non-zero value
        # 55 AA is the end of sector marker
        Regex(
            r"\xEB\x52\x90\x4E\x54\x46\x53\x20\x20\x20\x20[\x00-\xFF][\x01-\xFF][\x00-\xFF]{497}\x55\xAA"
        )
    ]

    C_DEFINITIONS = r"""
        typedef struct ntfs_boot {
            uint8 jmp_ins[3];	// jump instruction to boot code.
            char oem_id[8];	// OEM ID, equal "NTFS    "

            // BPB
            uint16 bytes_per_sector;
            uint8 sectors_per_clusters;
            uint8 unused1[7];
            uint8 media_descriptor;
            uint8 unused2[2];
            uint16 sectors_per_track;
            uint16 heads;
            uint32 hidden_sectors;
            uint8 unused3[4];

            // EBPB
            //uint8 bios_drive_num;	// 0x24: BIOS drive number =0x80.
            uint8 unused4[4];
            uint64 total_sectors;
            uint64 mft_cluster_number;
            uint64 mft_mirror_cluster_number;
            uint8 record_size;
            uint8 unused5[3];
            uint8 index_size;
            uint8 unused6[3];
            uint64 serial_num;
            uint32 checksum;

            uint8 bootstrap_code[426];
            uint16 boot_magic;
        } ntfs_boot_t;
    """

    HEADER_STRUCT = "ntfs_boot_t"

    EXTRACTOR = Command("7z", "x", "-x![SYSTEM]", "-y", "{inpath}", "-o{outdir}")

    DOC = HandlerDoc(
        name="NTFS",
        description="NTFS (New Technology File System) is a proprietary file system developed by Microsoft, featuring metadata support, advanced data structures, and journaling for reliability. It is commonly used in Windows operating systems for efficient storage and retrieval of files.",
        handler_type=HandlerType.FILESYSTEM,
        vendor="Microsoft",
        references=[
            Reference(
                title="NTFS Wikipedia",
                url="https://en.wikipedia.org/wiki/NTFS",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        header = self.parse_header(file)

        fsize = header.total_sectors * header.bytes_per_sector
        if fsize == 0:
            raise InvalidInputFormat("NTFS header with null disk size.")

        end_offset = start_offset + len(header) + fsize
        return ValidChunk(
            start_offset=start_offset,
            end_offset=end_offset,
        )



import abc
import contextlib
import multiprocessing as mp
import os
import queue
import signal
import sys
import threading
from multiprocessing.queues import JoinableQueue
from typing import Any, Callable, Union

from .logging import multiprocessing_breakpoint

mp.set_start_method("fork")


class PoolBase(abc.ABC):
    def __init__(self):
        with pools_lock:
            pools.add(self)

    @abc.abstractmethod
    def submit(self, args):
        pass

    @abc.abstractmethod
    def process_until_done(self):
        pass

    def start(self):
        pass

    def close(self, *, immediate=False):  # noqa: ARG002
        with pools_lock:
            pools.remove(self)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, _exc_value, _tb):
        self.close(immediate=exc_type is not None)


pools_lock = threading.Lock()
pools: set[PoolBase] = set()


class Queue(JoinableQueue):
    def is_empty(self) -> bool:
        """Check if all ``task_done`` has been called for all items.

        Based on ``multiprocessing.JoinableQueue.join``.
        """
        with self._cond:  # type: ignore
            return self._unfinished_tasks._semlock._is_zero()  # type: ignore  # noqa: SLF001


class _Sentinel:
    pass


_SENTINEL = _Sentinel


def _worker_process(handler, input_, output):
    # Creates a new process group, making sure no signals are
    # propagated from the main process to the worker processes.
    os.setpgrp()

    # Restore default signal handlers, otherwise workers would inherit
    # them from main process
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    sys.breakpointhook = multiprocessing_breakpoint
    while (args := input_.get()) is not _SENTINEL:
        result = handler(args)
        output.put(result)
    output.put(_SENTINEL)


class MultiPool(PoolBase):
    def __init__(
        self,
        process_num: int,
        handler: Callable[[Any], Any],
        *,
        result_callback: Callable[["MultiPool", Any], Any],
    ):
        super().__init__()
        if process_num <= 0:
            raise ValueError("At process_num must be greater than 0")

        self._running = False
        self._result_callback = result_callback
        self._input = Queue(ctx=mp.get_context())
        self._input.cancel_join_thread()
        self._output = mp.SimpleQueue()
        self._procs = [
            mp.Process(
                target=_worker_process,
                args=(handler, self._input, self._output),
            )
            for _ in range(process_num)
        ]
        self._tid = threading.get_native_id()

    def start(self):
        self._running = True
        for p in self._procs:
            p.start()

    def close(self, *, immediate=False):
        if not self._running:
            return
        self._running = False

        if immediate:
            self._terminate_workers()
        else:
            self._clear_input_queue()
            self._request_workers_to_quit()
            self._clear_output_queue()

        self._wait_for_workers_to_quit()
        super().close(immediate=immediate)

    def _terminate_workers(self):
        for proc in self._procs:
            proc.terminate()

        self._input.close()
        self._output.close()

    def _clear_input_queue(self):
        try:
            while True:
                self._input.get_nowait()
        except queue.Empty:
            pass

    def _request_workers_to_quit(self):
        for _ in self._procs:
            self._input.put(_SENTINEL)
        self._input.close()

    def _clear_output_queue(self):
        process_quit_count = 0
        process_num = len(self._procs)
        while process_quit_count < process_num:
            result = self._output.get()
            if result is _SENTINEL:
                process_quit_count += 1

    def _wait_for_workers_to_quit(self):
        for p in self._procs:
            p.join()

    def submit(self, args):
        if threading.get_native_id() != self._tid:
            raise RuntimeError(
                "Submit can only be called from the same "
                "thread/process where the pool is created"
            )
        self._input.put(args)

    def process_until_done(self):
        with contextlib.suppress(EOFError):
            while not self._input.is_empty():
                result = self._output.get()
                self._result_callback(self, result)
                self._input.task_done()


class SinglePool(PoolBase):
    def __init__(self, handler, *, result_callback):
        super().__init__()
        self._handler = handler
        self._result_callback = result_callback

    def submit(self, args):
        result = self._handler(args)
        self._result_callback(self, result)

    def process_until_done(self):
        pass


def make_pool(process_num, handler, result_callback) -> Union[SinglePool, MultiPool]:
    if process_num == 1:
        return SinglePool(handler=handler, result_callback=result_callback)

    return MultiPool(
        process_num=process_num,
        handler=handler,
        result_callback=result_callback,
    )


orig_signal_handlers = {}


def _on_terminate(signum, frame):
    pools_snapshot = list(pools)
    for pool in pools_snapshot:
        pool.close(immediate=True)

    if callable(orig_signal_handlers[signum]):
        orig_signal_handlers[signum](signum, frame)


orig_signal_handlers[signal.SIGTERM] = signal.signal(signal.SIGTERM, _on_terminate)
orig_signal_handlers[signal.SIGINT] = signal.signal(signal.SIGINT, _on_terminate)



import mmap
import shutil
from pathlib import Path

from structlog import get_logger

from unblob.file_utils import File
from unblob.handlers.archive.zip import ZIPHandler
from unblob.models import Extractor, HandlerDoc, HandlerType, HexString

logger = get_logger()

LOCAL_FILE_HEADER_ZIP = b"\x50\x4b\x03\x04"
LOCAL_FILE_HEADER_INSTAR = b"\x50\x4b\x03\x07"

EOCD_ZIP = b"\x50\x4b\x05\x06"
EOCD_INSTAR = b"\x50\x4b\x05\x09"

CD_FILE_HEADER_ZIP = b"\x50\x4b\x01\x02"
CD_FILE_HEADER_INSTAR = b"\x50\x4b\x01\x08"


class InstarHDExtractor(Extractor):
    def extract(self, inpath: Path, outdir: Path):
        outfile = outdir / "instar_hd.deobfuscated"
        shutil.copyfile(inpath, outfile)
        with File.from_path(outfile, access=mmap.ACCESS_WRITE) as mm:
            replacements = {
                LOCAL_FILE_HEADER_INSTAR: LOCAL_FILE_HEADER_ZIP,
                EOCD_INSTAR: EOCD_ZIP,
                CD_FILE_HEADER_INSTAR: CD_FILE_HEADER_ZIP,
            }
            for pattern, replacement in replacements.items():
                pos = mm.find(pattern)
                while pos != -1:
                    mm[pos : pos + len(pattern)] = replacement
                    pos = mm.find(pattern, pos + len(pattern))


class InstarHDHandler(ZIPHandler):
    NAME = "instar_hd"

    PATTERNS = [HexString("50 4b 03 07")]  # match on the modified zip header

    EXTRACTOR = InstarHDExtractor()

    EOCD_RECORD_HEADER = 0x9054B50

    DOC = HandlerDoc(
        name="Instar HD",
        description="Instar HD firmware files are modified ZIP archives with non-standard local file headers, central directory headers, and end-of-central-directory records. These modifications include custom magic bytes to differentiate them from standard ZIP files.",
        handler_type=HandlerType.ARCHIVE,
        vendor="Instar",
        references=[],
        limitations=[],
    )



import io
from pathlib import Path
from typing import Optional

from structlog import get_logger

from unblob.extractor import carve_chunk_to_file
from unblob.file_utils import Endian, File, InvalidInputFormat, StructParser, snull
from unblob.models import (
    Chunk,
    Extractor,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    ValidChunk,
)

logger = get_logger()

C_DEFINITIONS = r"""
    typedef struct bdl_toc_entry {
        uint64 offset;
        uint64 size;
    } bdl_toc_entry_t;

    typedef struct bdl_header {
        char magic[4];
        uint16 major;
        uint16 minor;
        uint32 toc_offset;
        char unknown[4];
        uint32 toc_entries;
        uint32 unknowns_2[3];
        char release[256];
        char brand[256];
        char device_id[256];
        char unknown_3[9];
        char version[256];
        char revision[256];
    } bdl_header_t;
"""


def is_valid_header(header) -> bool:
    if header.toc_offset == 0 or header.toc_entries == 0:
        return False
    try:
        snull(header.release).decode("utf-8")
        snull(header.brand).decode("utf-8")
        snull(header.device_id).decode("utf-8")
        snull(header.version).decode("utf-8")
        snull(header.revision).decode("utf-8")
    except UnicodeDecodeError:
        return False
    return True


class HPBDLExtractor(Extractor):
    def __init__(self):
        self._struct_parser = StructParser(C_DEFINITIONS)

    def extract(self, inpath: Path, outdir: Path):
        entries = []
        with File.from_path(inpath) as file:
            header = self._struct_parser.parse("bdl_header_t", file, Endian.LITTLE)
            file.seek(header.toc_offset, io.SEEK_SET)
            for i in range(header.toc_entries):
                entry = self._struct_parser.parse(
                    "bdl_toc_entry_t", file, Endian.LITTLE
                )
                entries.append(
                    (
                        outdir.joinpath(outdir.joinpath(Path(f"ipkg{i:03}"))),
                        Chunk(
                            start_offset=entry.offset,
                            end_offset=entry.offset + entry.size,
                        ),
                    )
                )

            for carve_path, chunk in entries:
                carve_chunk_to_file(
                    file=file,
                    chunk=chunk,
                    carve_path=carve_path,
                )


class HPBDLHandler(StructHandler):
    NAME = "bdl"

    PATTERNS = [HexString("69 62 64 6C 01 00 01 00")]

    C_DEFINITIONS = C_DEFINITIONS
    HEADER_STRUCT = "bdl_header_t"
    EXTRACTOR = HPBDLExtractor()

    DOC = HandlerDoc(
        name="HP BDL",
        description="The HP BDL format is a firmware archive containing a custom header and a table of contents that specifies offsets and sizes of embedded firmware components. It includes metadata such as release, brand, device ID, version, and revision.",
        handler_type=HandlerType.ARCHIVE,
        vendor="HP",
        references=[
            Reference(
                title="hpbdl",
                url="https://github.com/tylerwhall/hpbdl",
            )
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        header = self.parse_header(file, endian=Endian.LITTLE)

        if not is_valid_header(header):
            raise InvalidInputFormat("Invalid BDL header.")

        file.seek(start_offset + header.toc_offset, io.SEEK_SET)
        end_offset = -1
        for _ in range(header.toc_entries):
            entry = self._struct_parser.parse("bdl_toc_entry_t", file, Endian.LITTLE)
            end_offset = max(end_offset, start_offset + entry.offset + entry.size)

        return ValidChunk(start_offset=start_offset, end_offset=end_offset)



from typing import Protocol

from rich import progress
from rich.style import Style

from .models import TaskResult


class ProgressReporter(Protocol):
    def __enter__(self): ...

    def __exit__(self, _exc_type, _exc_value, _tb): ...

    def update(self, result: TaskResult): ...


class NullProgressReporter:
    def __enter__(self):
        pass

    def __exit__(self, _exc_type, _exc_value, _tb):
        pass

    def update(self, result: TaskResult):
        pass


class RichConsoleProgressReporter:
    def __init__(self):
        self._progress = progress.Progress(
            progress.TextColumn(
                "Extraction progress: {task.percentage:>3.0f}%",
                style=Style(color="#00FFC8"),
            ),
            progress.BarColumn(
                complete_style=Style(color="#00FFC8"), style=Style(color="#002060")
            ),
        )
        self._overall_progress_task = self._progress.add_task("Extraction progress:")

    def __enter__(self):
        self._progress.start()

    def __exit__(self, _exc_type, _exc_value, _tb):
        self._progress.remove_task(self._overall_progress_task)
        self._progress.stop()

    def update(self, result: TaskResult):
        if (total := self._progress.tasks[0].total) is not None:
            self._progress.update(
                self._overall_progress_task,
                advance=1,
                total=total + len(result.subtasks),
            )



import contextlib
import os
import tarfile
from pathlib import Path
from typing import Optional

from structlog import get_logger

from ...file_utils import OffsetFile, SeekError, decode_int, round_up, snull
from ...models import (
    Extractor,
    ExtractResult,
    File,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    Regex,
    StructHandler,
    ValidChunk,
)
from ._safe_tarfile import SafeTarFile

logger = get_logger()


BLOCK_SIZE = 512
END_OF_ARCHIVE_MARKER_SIZE = 2 * BLOCK_SIZE

MAGIC_OFFSET = 257

ZERO_BLOCK = bytes([0]) * BLOCK_SIZE


def _get_tar_end_offset(file: File, offset=0):
    file_with_offset = OffsetFile(file, offset)

    # First find the end of the last entry in the file
    last_offset = _get_end_of_last_tar_entry(file_with_offset)
    if last_offset == -1:
        return -1

    # Then find where the final zero blocks end
    return offset + _find_end_of_padding(file_with_offset, find_from=last_offset)


def _get_end_of_last_tar_entry(file) -> int:
    try:
        tf = tarfile.TarFile(mode="r", fileobj=file)
    except tarfile.TarError:
        return -1

    last_member = None

    try:
        for member in tf:
            last_member = member
    except (tarfile.TarError, SeekError):
        # recover what's already been parsed
        pass

    if last_member is None:
        return -1

    end_of_last_tar_entry = tf.offset
    try:
        file.seek(end_of_last_tar_entry)
    except SeekError:
        # last tar entry is truncated
        end_of_last_tar_entry = last_member.offset
        file.seek(end_of_last_tar_entry)

    return end_of_last_tar_entry


def _find_end_of_padding(file, *, find_from: int) -> int:
    find_from = round_up(find_from, BLOCK_SIZE)
    find_to = round_up(find_from + END_OF_ARCHIVE_MARKER_SIZE, tarfile.RECORDSIZE)

    max_padding_blocks = (find_to - find_from) // BLOCK_SIZE

    try:
        file.seek(find_from)
    except SeekError:
        # match to end of truncated file
        return file.seek(0, os.SEEK_END)

    for padding_blocks in range(max_padding_blocks):  # noqa: B007
        if file.read(BLOCK_SIZE) != ZERO_BLOCK:
            break
    else:
        padding_blocks = max_padding_blocks

    return find_from + padding_blocks * BLOCK_SIZE


class TarExtractor(Extractor):
    def extract(self, inpath: Path, outdir: Path):
        with contextlib.closing(SafeTarFile(inpath)) as tarfile:
            tarfile.extractall(outdir)  # noqa: S202 tarfile-unsafe-members
        return ExtractResult(reports=tarfile.reports)


class _TarHandler(StructHandler):
    NAME = "tar"

    PATTERNS = []

    C_DEFINITIONS = r"""
        typedef struct posix_header
        {                       /* byte offset */
            char name[100];     /*   0 */
            char mode[8];       /* 100 */
            char uid[8];        /* 108 */
            char gid[8];        /* 116 */
            char size[12];      /* 124 */
            char mtime[12];     /* 136 */
            char chksum[8];     /* 148 */
            char typeflag;      /* 156 */
            char linkname[100]; /* 157 */
            char magic[6];      /* 257 */
            char version[2];    /* 263 */
            char uname[32];     /* 265 */
            char gname[32];     /* 297 */
            char devmajor[8];   /* 329 */
            char devminor[8];   /* 337 */
            char prefix[155];   /* 345 */
                                /* 500 */
        } posix_header_t;
    """
    HEADER_STRUCT = "posix_header_t"

    EXTRACTOR = TarExtractor()

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        file.seek(start_offset)
        header = self.parse_header(file)
        header_size = snull(header.size)
        decode_int(header_size, 8)

        def signed_sum(octets) -> int:
            return sum(b if b < 128 else 256 - b for b in octets)

        if header.chksum[6:8] not in (b"\x00 ", b" \x00"):
            logger.debug(
                "Invalid checksum format",
                actual_last_2_bytes=header.chksum[6:8],
                handler=self.NAME,
                _verbosity=3,
            )
            return None
        checksum = decode_int(header.chksum[:6], 8)
        header_bytes_for_checksum = (
            file[start_offset : start_offset + 148]
            + b" " * 8  # chksum field is replaced with "blanks"
            + file[start_offset + 156 : start_offset + 257]
        )
        extended_header_bytes = file[start_offset + 257 : start_offset + 500]
        calculated_checksum_unsigned = sum(header_bytes_for_checksum)
        calculated_checksum_signed = signed_sum(header_bytes_for_checksum)
        checksums = (
            calculated_checksum_unsigned,
            calculated_checksum_unsigned + sum(extended_header_bytes),
            # signed is of historical interest, calculating for the extended header is not needed
            calculated_checksum_signed,
        )
        if checksum not in checksums:
            logger.error(
                "Tar header checksum mismatch", expected=str(checksum), actual=checksums
            )
            return None

        end_offset = _get_tar_end_offset(file, start_offset)
        if end_offset == -1:
            return None
        return ValidChunk(start_offset=start_offset, end_offset=end_offset)


class TarUstarHandler(_TarHandler):
    PATTERNS = [
        HexString("75 73 74 61 72 20 20 00"),
        HexString("75 73 74 61 72 00 30 30"),
    ]

    # Since the magic is at 257, we have to subtract that from the match offset
    # to get to the start of the file.
    PATTERN_MATCH_OFFSET = -MAGIC_OFFSET

    DOC = HandlerDoc(
        name="TAR (USTAR)",
        description="USTAR (Uniform Standard Tape Archive) tar files are extensions of the original tar format with additional metadata fields.",
        handler_type=HandlerType.ARCHIVE,
        vendor=None,
        references=[
            Reference(
                title="USTAR Format Documentation",
                url="https://en.wikipedia.org/wiki/Tar_(computing)#USTAR_format",
            ),
            Reference(
                title="POSIX Tar Format Specification",
                url="https://pubs.opengroup.org/onlinepubs/9699919799/utilities/pax.html",
            ),
        ],
        limitations=[],
    )


def _re_frame(regexp: str):
    """Wrap regexp to ensure its integrity from concatenation.

    E.g.: when the regex
      a|b
    is naively appended by regex c, the result
      a|bc
    will not match "ac", while
      (a|b)c
    will match "ac" as intended.
    """
    return f"({regexp})"


def _re_alternatives(regexps):
    return _re_frame("|".join(_re_frame(regexp) for regexp in regexps))


def _padded_field(re_content_char, size, leftpad_re=" ", rightpad_re=r"[ \0x00]"):
    field_regexes = []

    for padsize in range(size):
        content_re = f"{re_content_char}{{{size - padsize}}}"

        for leftpadsize in range(padsize + 1):
            rightpadsize = padsize - leftpadsize

            left_re = f"{leftpad_re}{{{leftpadsize}}}" if leftpadsize else ""
            right_re = f"{rightpad_re}{{{rightpadsize}}}" if rightpadsize else ""

            field_regexes.append(f"{left_re}{content_re}{right_re}")

    return _re_alternatives(field_regexes)


class TarUnixHandler(_TarHandler):
    PATTERNS = [
        Regex(
            r""
            #  (pattern would be too big)   char name[100]
            + _padded_field(r"[0-7]", 8)  # char mode[8]
            + _padded_field(r"[0-7]", 8)  # char uid[8]
            + _padded_field(r"[0-7]", 8)  # char gid[8]
            + _padded_field(r"[0-7]", 12)  # char size[12]
            + _padded_field(r"[0-7]", 12)  # char mtime[12]
            + _padded_field(r"[0-7]", 8)  # char chksum[8]
            + r"[0-7\x00]"  # char typeflag[1] - no extensions
        ),
    ]
    PATTERN_MATCH_OFFSET = -100  # go back to beginning of skipped name

    DOC = HandlerDoc(
        name="TAR (Unix)",
        description="Unix tar files are a widely used archive format for storing files and directories with metadata.",
        handler_type=HandlerType.ARCHIVE,
        vendor=None,
        references=[
            Reference(
                title="Unix Tar Format Documentation",
                url="https://en.wikipedia.org/wiki/Tar_(computing)",
            ),
            Reference(
                title="GNU Tar Manual",
                url="https://www.gnu.org/software/tar/manual/",
            ),
        ],
        limitations=[],
    )



import hashlib
import io
from pathlib import Path
from typing import Optional

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from structlog import get_logger

from unblob.file_utils import File, InvalidInputFormat
from unblob.models import (
    Endian,
    Extractor,
    HandlerDoc,
    HandlerType,
    HexString,
    Reference,
    StructHandler,
    StructParser,
    ValidChunk,
)

logger = get_logger()

C_DEFINITIONS = r"""

    typedef struct dlink_shrs_header {
        char magic[4];                   /* SHRS */
        uint32 file_size;                /* Length of decrypted firmware in bytes */
        uint32 file_size_no_padding;     /* Length of decrypted firmware - padding */
        char iv[16];                     /* Length of AES 128 cbc IV */
        char decrypted_key_digest[64];          /* SHA512 64 byte message digest of decrypted firmware + key */
        char decrypted_digest[64];              /* SHA512 64 byte message digest of decrypted firmware */
        char encrypted_digest[64];              /* SHA512 64 byte message digest of encrypted firmware */
        char unused[512];                /* 512 unused NULL bytes (0xdc to 0x2dc) */
        char signature_1[512];           /* 512 byte Signature 1 */
        char signature_2[512];           /* 512 byte Signature 2 */
    } dlink_header_t;
"""

KEY = bytes.fromhex("c05fbf1936c99429ce2a0781f08d6ad8")
sha512 = hashlib.sha512()


class SHRSExtractor(Extractor):
    def __init__(self):
        self._struct_parser = StructParser(C_DEFINITIONS)

    def extract(self, inpath: Path, outdir: Path):
        outpath = outdir.joinpath(f"{inpath.name}.decrypted")
        with File.from_path(inpath) as file:
            header = self._struct_parser.parse(
                "dlink_header_t", file, endian=Endian.BIG
            )
            cipher = Cipher(algorithms.AES(KEY), modes.CBC(header.iv))
            decryptor = cipher.decryptor()
            outfile = outpath.open("wb")
            ciphertext = file.read(1024)
            while ciphertext and len(ciphertext) % 16 == 0:
                plaintext = decryptor.update(ciphertext)
                outfile.write(plaintext)
                ciphertext = file.read(1024)
        outfile.write(decryptor.finalize())
        outfile.close()


class SHRSHandler(StructHandler):
    NAME = "shrs"

    PATTERNS = [HexString("53 48 52 53")]

    C_DEFINITIONS = C_DEFINITIONS
    HEADER_STRUCT = "dlink_header_t"
    EXTRACTOR = SHRSExtractor()

    DOC = HandlerDoc(
        name="D-Link SHRS",
        description="SHRS is a D-Link firmware format with a custom header containing metadata, SHA-512 digests, and AES-CBC encryption parameters. The firmware data is encrypted using a fixed key and IV stored in the header.",
        handler_type=HandlerType.ARCHIVE,
        vendor="D-Link",
        references=[
            Reference(
                title="Breaking the D-Link DIR3060 Firmware Encryption - Recon - Part 1",
                url="https://0x00sec.org/t/breaking-the-d-link-dir3060-firmware-encryption-recon-part-1/21943",  # Replace with actual reference if available
            )
        ],
        limitations=[],
    )

    def is_valid_header(self, header, file: File) -> bool:
        if header.file_size < len(header):
            return False
        # we're exactly past the header, we compute the digest
        digest = hashlib.sha512(file.read(header.file_size_no_padding)).digest()
        # we seek back to where we were
        file.seek(-header.file_size_no_padding, io.SEEK_CUR)
        return digest == header.encrypted_digest

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        header = self.parse_header(file, endian=Endian.BIG)

        if not self.is_valid_header(header, file):
            raise InvalidInputFormat("Invalid SHRS header.")

        cipher = Cipher(algorithms.AES(KEY), modes.CBC(header.iv))
        decryptor = cipher.decryptor()
        read_bytes = 0

        while read_bytes < header.file_size:
            read_bytes += len(decryptor.update(file.read(1024)))
        end_offset = file.tell()

        return ValidChunk(start_offset=start_offset, end_offset=end_offset)



import hashlib
import stat
import traceback
from enum import Enum
from pathlib import Path
from typing import Optional, Union, final

import attrs


@attrs.define(kw_only=True, frozen=True)
class Report:
    """A common base class for different reports."""

    def __attrs_post_init__(self):
        for field in attrs.fields(type(self)):
            value = getattr(self, field.name)
            if isinstance(value, int):
                object.__setattr__(self, field.name, int(value))

    def asdict(self) -> dict:
        return attrs.asdict(self)


class Severity(Enum):
    """Represents possible problems encountered during execution."""

    ERROR = "ERROR"
    WARNING = "WARNING"


@attrs.define(kw_only=True, frozen=True)
class ErrorReport(Report):
    severity: Severity


def _convert_exception_to_str(obj: Union[str, Exception]) -> str:
    if isinstance(obj, str):
        return obj
    if isinstance(obj, Exception):
        e: Exception = obj
        return "".join(traceback.format_exception(type(e), e, e.__traceback__))
    raise ValueError("Invalid exception object", obj)


@attrs.define(kw_only=True, frozen=True)
class UnknownError(ErrorReport):
    """Describes an exception raised during file processing."""

    severity: Severity = attrs.field(default=Severity.ERROR)
    exception: Union[str, Exception] = attrs.field(  # pyright: ignore[reportGeneralTypeIssues]
        converter=_convert_exception_to_str
    )
    """Exceptions are also formatted at construct time.

    `attrs` is not integrated enough with type checker/LSP provider `pyright` to include converters.

    See: https://www.attrs.org/en/stable/types.html#pyright
    """


@attrs.define(kw_only=True, frozen=True)
class CalculateChunkExceptionReport(UnknownError):
    """Describes an exception raised during calculate_chunk execution."""

    start_offset: int
    # Stored in `str` rather than `Handler`, because the pickle picks ups structs from `C_DEFINITIONS`
    handler: str


@attrs.define(kw_only=True, frozen=True)
class CalculateMultiFileExceptionReport(UnknownError):
    """Describes an exception raised during calculate_chunk execution."""

    path: Path
    # Stored in `str` rather than `Handler`, because the pickle picks ups structs from `C_DEFINITIONS`
    handler: str


@attrs.define(kw_only=True, frozen=True)
class ExtractCommandFailedReport(ErrorReport):
    """Describes an error when failed to run the extraction command."""

    severity: Severity = Severity.WARNING
    command: str
    stdout: bytes
    stderr: bytes
    exit_code: int


@attrs.define(kw_only=True, frozen=True)
class OutputDirectoryExistsReport(ErrorReport):
    severity: Severity = Severity.ERROR
    path: Path


@attrs.define(kw_only=True, frozen=True)
class ExtractorDependencyNotFoundReport(ErrorReport):
    """Describes an error when the dependency of an extractor doesn't exist."""

    severity: Severity = Severity.ERROR
    dependencies: list[str]


@attrs.define(kw_only=True, frozen=True)
class ExtractorTimedOut(ErrorReport):
    """Describes an error when the extractor execution timed out."""

    severity: Severity = Severity.ERROR
    cmd: str
    timeout: float


@attrs.define(kw_only=True, frozen=True)
class MaliciousSymlinkRemoved(ErrorReport):
    """Describes an error when malicious symlinks have been removed from disk."""

    severity: Severity = Severity.WARNING
    link: str
    target: str


@attrs.define(kw_only=True, frozen=True)
class MultiFileCollisionReport(ErrorReport):
    """Describes an error when MultiFiles collide on the same file."""

    severity: Severity = Severity.ERROR
    paths: set[Path]
    handler: str


@attrs.define(kw_only=True, frozen=True)
class StatReport(Report):
    path: Path
    size: int
    is_dir: bool
    is_file: bool
    is_link: bool
    link_target: Optional[Path]

    @classmethod
    def from_path(cls, path: Path):
        st = path.lstat()
        mode = st.st_mode
        try:
            link_target = Path.readlink(path)
        except OSError:
            link_target = None

        return cls(
            path=path,
            size=st.st_size,
            is_dir=stat.S_ISDIR(mode),
            is_file=stat.S_ISREG(mode),
            is_link=stat.S_ISLNK(mode),
            link_target=link_target,
        )


@attrs.define(kw_only=True, frozen=True)
class HashReport(Report):
    md5: str
    sha1: str
    sha256: str

    @classmethod
    def from_path(cls, path: Path):
        chunk_size = 1024 * 64
        md5 = hashlib.md5()  # noqa: S324
        sha1 = hashlib.sha1()  # noqa: S324
        sha256 = hashlib.sha256()

        with path.open("rb") as f:
            while chunk := f.read(chunk_size):
                md5.update(chunk)
                sha1.update(chunk)
                sha256.update(chunk)

        return cls(
            md5=md5.hexdigest(),
            sha1=sha1.hexdigest(),
            sha256=sha256.hexdigest(),
        )


@attrs.define(kw_only=True, frozen=True)
class FileMagicReport(Report):
    magic: str
    mime_type: str


@attrs.define(kw_only=True, frozen=True)
class RandomnessMeasurements:
    percentages: list[float]
    block_size: int
    mean: float

    @property
    def highest(self):
        return max(self.percentages)

    @property
    def lowest(self):
        return min(self.percentages)


@attrs.define(kw_only=True, frozen=True)
class RandomnessReport(Report):
    shannon: RandomnessMeasurements
    chi_square: RandomnessMeasurements


@final
@attrs.define(kw_only=True, frozen=True)
class ChunkReport(Report):
    id: str
    handler_name: str
    start_offset: int
    end_offset: int
    size: int
    is_encrypted: bool
    extraction_reports: list[Report]


@final
@attrs.define(kw_only=True, frozen=True)
class UnknownChunkReport(Report):
    id: str
    start_offset: int
    end_offset: int
    size: int
    randomness: Optional[RandomnessReport]


@attrs.define(kw_only=True, frozen=True)
class CarveDirectoryReport(Report):
    carve_dir: Path


@final
@attrs.define(kw_only=True, frozen=True)
class MultiFileReport(Report):
    id: str
    handler_name: str
    name: str
    paths: list[Path]
    extraction_reports: list[Report]


@attrs.define(kw_only=True, frozen=True)
class ExtractionProblem(Report):
    """A non-fatal problem discovered during extraction.

    A report like this still means, that the extraction was successful,
    but there were problems that got resolved.
    The output is expected to be complete, with the exception of
    the reported path.

    Examples
    --------
    - duplicate entries for certain archive formats (tar, zip)
    - unsafe symlinks pointing outside of extraction directory

    """

    problem: str
    resolution: str
    path: Optional[str] = None

    @property
    def log_msg(self):
        return f"{self.problem} {self.resolution}"

    def log_with(self, logger):
        logger.warning(self.log_msg, path=self.path)


@attrs.define(kw_only=True, frozen=True)
class PathTraversalProblem(ExtractionProblem):
    extraction_path: str

    def log_with(self, logger):
        logger.warning(
            self.log_msg,
            path=self.path,
            extraction_path=self.extraction_path,
        )


@attrs.define(kw_only=True, frozen=True)
class LinkExtractionProblem(ExtractionProblem):
    link_path: str

    def log_with(self, logger):
        logger.warning(self.log_msg, path=self.path, link_path=self.link_path)


@attrs.define(kw_only=True, frozen=True)
class SpecialFileExtractionProblem(ExtractionProblem):
    mode: int
    device: int

    def log_with(self, logger):
        logger.warning(self.log_msg, path=self.path, mode=self.mode, device=self.device)



import shutil

import attrs

from .models import DirectoryHandlers, Handlers


@attrs.define
class Dependency:
    command: str
    is_installed: bool


INSTALLED = "✓"
NOT_INSTALLED = "✗"


def get_dependencies(
    handlers: Handlers, dir_handlers: DirectoryHandlers
) -> list[Dependency]:
    all_commands = set()
    for handler in handlers:
        commands = handler.get_dependencies()
        all_commands.update(commands)
    for handler in dir_handlers:
        commands = handler.get_dependencies()
        all_commands.update(commands)
    rv = []
    for command in sorted(all_commands):
        is_installed = shutil.which(command) is not None
        rv.append(Dependency(command=command, is_installed=is_installed))
    return rv


def pretty_format_dependencies(dependencies: list[Dependency]) -> str:
    longest_key_length = max(len(dep.command) for dep in dependencies)
    lines = ["The following executables found installed, which are needed by unblob:"]
    for dependency in dependencies:
        mark = INSTALLED if dependency.is_installed else NOT_INSTALLED
        lines.append(f"    {dependency.command:<{longest_key_length}}    {mark}")
    return "\n".join(lines)



import io
from typing import Optional

from structlog import get_logger

from unblob.extractors.command import Command

from ....file_utils import Endian
from ....models import (
    File,
    HandlerDoc,
    HandlerType,
    Reference,
    Regex,
    StructHandler,
    ValidChunk,
)

logger = get_logger()

CHUNK_TYPE_RAW = 0xCAC1
CHUNK_TYPE_FILL = 0xCAC2
CHUNK_TYPE_DONT_CARE = 0xCAC3
CHUNK_TYPE_CRC32 = 0xCAC4

VALID_CHUNK_TYPES = [
    CHUNK_TYPE_RAW,
    CHUNK_TYPE_FILL,
    CHUNK_TYPE_DONT_CARE,
    CHUNK_TYPE_CRC32,
]


class SparseHandler(StructHandler):
    NAME = "sparse"

    # magic (0xed26ff3a)
    # major version (0x1)
    # minor version (any)
    # file header size (0x1C in v1.0)
    # chunk header size (0XC in v1.0)
    PATTERNS = [Regex(r"\x3A\xFF\x26\xED\x01\x00[\x00-\xFF]{2}\x1C\x00\x0C\x00")]

    C_DEFINITIONS = r"""
        typedef struct sparse_header {
            uint32 magic;          /* 0xed26ff3a */
            uint16 major_version;  /* (0x1) - reject images with higher major versions */
            uint16 minor_version;  /* (0x0) - allow images with higer minor versions */
            uint16 file_hdr_sz;    /* 28 bytes for first revision of the file format */
            uint16 chunk_hdr_sz;   /* 12 bytes for first revision of the file format */
            uint32 blk_sz;         /* block size in bytes, must be a multiple of 4 (4096) */
            uint32 total_blks;     /* total blocks in the non-sparse output image */
            uint32 total_chunks;   /* total chunks in the sparse input image */
            uint32 image_checksum; /* CRC32 checksum of the original data, counting "don't care" */
                                    /* as 0. Standard 802.3 polynomial, use a Public Domain */
                                    /* table implementation */
        } sparse_header_t;

        typedef struct chunk_header {
            uint16 chunk_type; /* 0xCAC1 -> raw; 0xCAC2 -> fill; 0xCAC3 -> don't care */
            uint16 reserved1;
            uint32 chunk_sz; /* in blocks in output image */
            uint32 total_sz; /* in bytes of chunk input file including chunk header and data */
        } chunk_header_t;
    """
    HEADER_STRUCT = "sparse_header_t"

    EXTRACTOR = Command("simg2img", "{inpath}", "{outdir}/raw.image")

    DOC = HandlerDoc(
        name="Android Sparse",
        description="Android sparse images are a file format used to efficiently store disk images by representing sequences of zero blocks compactly. The format includes a file header, followed by chunk headers and data, with support for raw, fill, and 'don't care' chunks.",
        handler_type=HandlerType.FILESYSTEM,
        vendor="Google",
        references=[
            Reference(
                title="Android Sparse Image Format Documentation",
                url="https://formats.kaitai.io/android_sparse/",
            ),
            Reference(
                title="simg2img Tool",
                url="https://github.com/anestisb/android-simg2img",
            ),
        ],
        limitations=[],
    )

    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        header = self.parse_header(file, Endian.LITTLE)

        count = 0
        while count < header.total_chunks:
            chunk_header = self.cparser_le.chunk_header_t(file)
            if chunk_header.chunk_type not in VALID_CHUNK_TYPES:
                logger.warning("Invalid chunk type in Android sparse image. Aborting.")
                return None
            file.seek(chunk_header.total_sz - len(chunk_header), io.SEEK_CUR)
            count += 1

        return ValidChunk(
            start_offset=start_offset,
            end_offset=file.tell(),
        )



import abc
import dataclasses
import itertools
import json
from collections.abc import Iterable
from enum import Enum
from pathlib import Path
from typing import Generic, Optional, TypeVar, Union

import attrs
from structlog import get_logger

from .file_utils import Endian, File, InvalidInputFormat, StructParser
from .identifiers import new_id
from .parser import hexstring2regex
from .report import (
    CarveDirectoryReport,
    ChunkReport,
    ErrorReport,
    MultiFileReport,
    RandomnessReport,
    Report,
    UnknownChunkReport,
)

logger = get_logger()

# The state transitions are:
#
# file ──► pattern match ──► ValidChunk
#


class HandlerType(Enum):
    ARCHIVE = "Archive"
    COMPRESSION = "Compression"
    FILESYSTEM = "FileSystem"
    EXECUTABLE = "Executable"
    BAREMETAL = "Baremetal"
    BOOTLOADER = "Bootloader"
    ENCRYPTION = "Encryption"


@dataclasses.dataclass(frozen=True)
class Reference:
    title: str
    url: str


@dataclasses.dataclass
class HandlerDoc:
    name: str
    description: Union[str, None]
    vendor: Union[str, None]
    references: list[Reference]
    limitations: list[str]
    handler_type: HandlerType
    fully_supported: bool = dataclasses.field(init=False)

    def __post_init__(self):
        self.fully_supported = len(self.limitations) == 0


@attrs.define(frozen=True)
class Task:
    path: Path
    depth: int
    blob_id: str
    is_multi_file: bool = attrs.field(default=False)


@attrs.define
class Blob:
    id: str = attrs.field(
        factory=new_id,
    )


@attrs.define
class Chunk(Blob):
    """File chunk, have start and end offset, but still can be invalid.

    For an array ``b``, a chunk ``c`` represents the slice:
    ::

        b[c.start_offset:c.end_offset]
    """

    start_offset: int = attrs.field(kw_only=True)
    """The index of the first byte of the chunk"""

    end_offset: int = attrs.field(kw_only=True)
    """The index of the first byte after the end of the chunk"""

    file: Optional[File] = None

    def __attrs_post_init__(self):
        if self.start_offset < 0 or self.end_offset < 0:
            raise InvalidInputFormat(f"Chunk has negative offset: {self}")
        if self.start_offset >= self.end_offset:
            raise InvalidInputFormat(
                f"Chunk has higher start_offset than end_offset: {self}"
            )

    @property
    def size(self) -> int:
        return self.end_offset - self.start_offset

    @property
    def range_hex(self) -> str:
        return f"0x{self.start_offset:x}-0x{self.end_offset:x}"

    @property
    def is_whole_file(self):
        assert self.file
        return self.start_offset == 0 and self.end_offset == self.file.size()

    def contains(self, other: "Chunk") -> bool:
        return (
            self.start_offset < other.start_offset
            and self.end_offset >= other.end_offset
        ) or (
            self.start_offset <= other.start_offset
            and self.end_offset > other.end_offset
        )

    def contains_offset(self, offset: int) -> bool:
        return self.start_offset <= offset < self.end_offset

    def __repr__(self) -> str:
        return self.range_hex


@attrs.define(repr=False)
class ValidChunk(Chunk):
    """Known to be valid chunk of a File, can be extracted with an external program."""

    handler: "Handler" = attrs.field(init=False, eq=False)
    is_encrypted: bool = attrs.field(default=False)

    def extract(self, inpath: Path, outdir: Path) -> Optional["ExtractResult"]:
        if self.is_encrypted:
            logger.warning(
                "Encrypted file is not extracted",
                path=inpath,
                chunk=self,
            )
            raise ExtractError

        return self.handler.extract(inpath, outdir)

    def as_report(self, extraction_reports: list[Report]) -> ChunkReport:
        return ChunkReport(
            id=self.id,
            start_offset=self.start_offset,
            end_offset=self.end_offset,
            size=self.size,
            handler_name=self.handler.NAME,
            is_encrypted=self.is_encrypted,
            extraction_reports=extraction_reports,
        )


@attrs.define(repr=False)
class UnknownChunk(Chunk):
    r"""Gaps between valid chunks or otherwise unknown chunks.

    Important for manual analysis, and analytical certainty: for example
    randomness, other chunks inside it, metadata, etc.

    These are not extracted, just logged for information purposes and further analysis,
    like most common bytes (like \x00 and \xFF), ASCII strings, high randomness, etc.
    """

    def as_report(self, randomness: Optional[RandomnessReport]) -> UnknownChunkReport:
        return UnknownChunkReport(
            id=self.id,
            start_offset=self.start_offset,
            end_offset=self.end_offset,
            size=self.size,
            randomness=randomness,
        )


@attrs.define(repr=False)
class PaddingChunk(Chunk):
    r"""Gaps between valid chunks or otherwise unknown chunks.

    Important for manual analysis, and analytical certanity: for example
    randomness, other chunks inside it, metadata, etc.
    """

    def as_report(
        self,
        randomness: Optional[RandomnessReport],  #   noqa: ARG002
    ) -> ChunkReport:
        return ChunkReport(
            id=self.id,
            start_offset=self.start_offset,
            end_offset=self.end_offset,
            size=self.size,
            is_encrypted=False,
            handler_name="padding",
            extraction_reports=[],
        )


@attrs.define
class MultiFile(Blob):
    name: str = attrs.field(kw_only=True)
    paths: list[Path] = attrs.field(kw_only=True)

    handler: "DirectoryHandler" = attrs.field(init=False, eq=False)

    def extract(self, outdir: Path) -> Optional["ExtractResult"]:
        return self.handler.extract(self.paths, outdir)

    def as_report(self, extraction_reports: list[Report]) -> MultiFileReport:
        return MultiFileReport(
            id=self.id,
            name=self.name,
            paths=self.paths,
            handler_name=self.handler.NAME,
            extraction_reports=extraction_reports,
        )


ReportType = TypeVar("ReportType", bound=Report)


@attrs.define
class TaskResult:
    task: Task
    reports: list[Report] = attrs.field(factory=list)
    subtasks: list[Task] = attrs.field(factory=list)

    def add_report(self, report: Report):
        self.reports.append(report)

    def add_subtask(self, task: Task):
        self.subtasks.append(task)

    def filter_reports(self, report_class: type[ReportType]) -> list[ReportType]:
        return [report for report in self.reports if isinstance(report, report_class)]


@attrs.define
class ProcessResult:
    results: list[TaskResult] = attrs.field(factory=list)

    @property
    def errors(self) -> list[ErrorReport]:
        reports = itertools.chain.from_iterable(r.reports for r in self.results)
        interesting_reports = (
            r for r in reports if isinstance(r, (ErrorReport, ChunkReport))
        )
        errors = []
        for report in interesting_reports:
            if isinstance(report, ErrorReport):
                errors.append(report)
            else:
                errors.extend(
                    r for r in report.extraction_reports if isinstance(r, ErrorReport)
                )
        return errors

    def register(self, result: TaskResult):
        self.results.append(result)

    def to_json(self, indent="  "):
        return to_json(self.results, indent=indent)

    def get_output_dir(self) -> Optional[Path]:
        try:
            top_result = self.results[0]
            if carves := top_result.filter_reports(CarveDirectoryReport):
                # we have a top level carve
                return carves[0].carve_dir

            # we either have an extraction,
            # and the extract directory registered as subtask
            return top_result.subtasks[0].path
        except IndexError:
            # or no extraction
            return None


class _JSONEncoder(json.JSONEncoder):
    def default(self, o):
        obj = o
        if attrs.has(type(obj)):
            extend_attr_output = True
            attr_output = attrs.asdict(obj, recurse=not extend_attr_output)
            attr_output["__typename__"] = obj.__class__.__name__
            return attr_output

        if isinstance(obj, Enum):
            return obj.name

        if isinstance(obj, Path):
            return str(obj)

        if isinstance(obj, bytes):
            try:
                return obj.decode()
            except UnicodeDecodeError:
                return str(obj)

        logger.error("JSONEncoder met a non-JSON encodable value", obj=obj)
        # the usual fail path of custom JSONEncoders is to call the parent and let it fail
        #     return json.JSONEncoder.default(self, obj)
        # instead of failing, just return something usable
        return f"Non-JSON encodable value: {obj}"


def to_json(obj, indent="  ") -> str:
    """Encode any UnBlob object as a serialized JSON."""
    return json.dumps(obj, cls=_JSONEncoder, indent=indent)


class ExtractError(Exception):
    """There was an error during extraction."""

    def __init__(self, *reports: Report):
        super().__init__()
        self.reports: tuple[Report, ...] = reports


@attrs.define(kw_only=True)
class ExtractResult:
    reports: list[Report]


class Extractor(abc.ABC):
    def get_dependencies(self) -> list[str]:
        """Return the external command dependencies."""
        return []

    @abc.abstractmethod
    def extract(self, inpath: Path, outdir: Path) -> Optional[ExtractResult]:
        """Extract the carved out chunk.

        Raises ExtractError on failure.
        """


class DirectoryExtractor(abc.ABC):
    def get_dependencies(self) -> list[str]:
        """Return the external command dependencies."""
        return []

    @abc.abstractmethod
    def extract(self, paths: list[Path], outdir: Path) -> Optional[ExtractResult]:
        """Extract from a multi file path list.

        Raises ExtractError on failure.
        """


class Pattern(str):
    def as_regex(self) -> bytes:
        raise NotImplementedError


class HexString(Pattern):
    """Hex string can be a YARA rule like hexadecimal string.

    It is useful to simplify defining binary strings using hex
    encoding, wild-cards, jumps and alternatives.  Hexstrings are
    convereted to hyperscan compatible PCRE regex.

    See YARA & Hyperscan documentation for more details:

        - https://yara.readthedocs.io/en/stable/writingrules.html#hexadecimal-strings

        - https://intel.github.io/hyperscan/dev-reference/compilation.html#pattern-support

    You can specify the following:

        - normal bytes using hexadecimals: 01 de ad co de ff

        - wild-cards can match single bytes and can be mixed with
          normal hex: 01 ??  02

        - wild-cards can also match first and second nibles: 0?  ?0

        - jumps can be specified for multiple wildcard bytes: [3]
          [2-5]

        - alternatives can be specified as well: ( 01 02 | 03 04 ) The
          above can be combined and alternatives nested: 01 02 ( 03 04
          | (0?  | 03 | ?0) | 05 ??  ) 06

    Single line comments can be specified using //

    We do NOT support the following YARA syntax:

        - comments using /* */ notation

        - infinite jumps: [-]

        - unbounded jumps: [3-] or [-4] (use [0-4] instead)
    """

    def as_regex(self) -> bytes:
        return hexstring2regex(self)


class Regex(Pattern):
    """Byte PCRE regex.

    See hyperscan documentation for more details:
    https://intel.github.io/hyperscan/dev-reference/compilation.html#pattern-support.
    """

    def as_regex(self) -> bytes:
        return self.encode()


class DirectoryPattern:
    def get_files(self, directory: Path) -> Iterable[Path]:
        raise NotImplementedError


class Glob(DirectoryPattern):
    def __init__(self, pattern):
        self._pattern = pattern

    def get_files(self, directory: Path) -> Iterable[Path]:
        return directory.glob(self._pattern)


class SingleFile(DirectoryPattern):
    def __init__(self, filename):
        self._filename = filename

    def get_files(self, directory: Path) -> Iterable[Path]:
        path = directory / self._filename
        return [path] if path.exists() else []


class DirectoryHandler(abc.ABC):
    """A directory type handler is responsible for searching, validating and "unblobbing" files from multiple files in a directory."""

    NAME: str

    EXTRACTOR: DirectoryExtractor

    PATTERN: DirectoryPattern

    DOC: HandlerDoc

    @classmethod
    def get_dependencies(cls):
        """Return external command dependencies needed for this handler to work."""
        if cls.EXTRACTOR:
            return cls.EXTRACTOR.get_dependencies()
        return []

    @abc.abstractmethod
    def calculate_multifile(self, file: Path) -> Optional[MultiFile]:
        """Calculate the MultiFile in a directory, using a file matched by the pattern as a starting point."""

    def extract(self, paths: list[Path], outdir: Path) -> Optional[ExtractResult]:
        if self.EXTRACTOR is None:
            logger.debug("Skipping file: no extractor.", paths=paths)
            raise ExtractError

        # We only extract every blob once, it's a mistake to extract the same blob again
        outdir.mkdir(parents=True, exist_ok=False)

        return self.EXTRACTOR.extract(paths, outdir)


TExtractor = TypeVar("TExtractor", bound=Union[None, Extractor])


class Handler(abc.ABC, Generic[TExtractor]):
    """A file type handler is responsible for searching, validating and "unblobbing" files from Blobs."""

    NAME: str
    PATTERNS: list[Pattern]
    # We need this, because not every match reflects the actual start
    # (e.g. tar magic is in the middle of the header)
    PATTERN_MATCH_OFFSET: int = 0

    EXTRACTOR: TExtractor

    DOC: HandlerDoc

    @classmethod
    def get_dependencies(cls):
        """Return external command dependencies needed for this handler to work."""
        if cls.EXTRACTOR is not None:
            return cls.EXTRACTOR.get_dependencies()
        return []

    @abc.abstractmethod
    def calculate_chunk(self, file: File, start_offset: int) -> Optional[ValidChunk]:
        """Calculate the Chunk offsets from the File and the file type headers."""

    def extract(self, inpath: Path, outdir: Path) -> Optional[ExtractResult]:
        if self.EXTRACTOR is None:
            logger.debug("Skipping file: no extractor.", path=inpath)
            raise ExtractError

        # We only extract every blob once, it's a mistake to extract the same blob again
        outdir.mkdir(parents=True, exist_ok=False)

        return self.EXTRACTOR.extract(inpath, outdir)


class StructHandler(Handler):
    C_DEFINITIONS: str
    # A struct from the C_DEFINITIONS used to parse the file's header
    HEADER_STRUCT: str

    def __init__(self):
        self._struct_parser = StructParser(self.C_DEFINITIONS)

    @property
    def cparser_le(self):
        return self._struct_parser.cparser_le

    @property
    def cparser_be(self):
        return self._struct_parser.cparser_be

    def parse_header(self, file: File, endian=Endian.LITTLE):
        header = self._struct_parser.parse(self.HEADER_STRUCT, file, endian)
        logger.debug("Header parsed", header=header, _verbosity=3)
        return header


Handlers = tuple[type[Handler], ...]
DirectoryHandlers = tuple[type[DirectoryHandler], ...]



from typing import Union

from .models import HandlerDoc, Reference

_HANDLER_DOC_MARKDOWN_TEMPLATE = """## {name}

!!! {support_info}

    === "Description"

{description}

        ---

        - **Handler type:** {type}
        {vendor}

    === "References"

{references}

{limitations}
"""


def _make_paragraph(lines: Union[str, None, list[str]]):  # noqa: C901
    if not lines:
        return ""

    if not isinstance(lines, list):
        lines = lines.splitlines()

    def _starting_enumeration_needs_newline():
        # Enumerations blocks needs an empty line before and after the blocks
        if not line.startswith("-"):
            return False
        if not formatted_lines:
            return False
        previous_line = formatted_lines[-1].strip()
        return previous_line != "" and not previous_line.startswith("-")

    formatted_lines = []
    for line in [line.strip() for line in lines]:
        if line:
            if _starting_enumeration_needs_newline():
                formatted_lines.append("")
            leading_space_count = 4 if line.startswith("===") else 8
            formatted_lines.append(" " * leading_space_count + line)
        else:
            formatted_lines.append(line)
    return "\n".join(formatted_lines)


def _make_url(*, text: str, link: str) -> str:
    return f'[{text}]({link}){{ target="_blank" }}'


def make_references(references: list[Reference]):
    if not references:
        return ""

    return _make_paragraph(
        [
            *[f"- {_make_url(text=ref.title, link=ref.url)}" for ref in references],
        ]
    )


def make_limitations(limitations: list[str]) -> str:
    if not limitations:
        return ""

    return '    === "Limitations"\n\n' + _make_paragraph(
        [
            *[f"- {limitation}" for limitation in limitations],
        ]
    )


def generate_markdown(doc: HandlerDoc) -> str:
    return _HANDLER_DOC_MARKDOWN_TEMPLATE.format(
        name=doc.name,
        type=doc.handler_type.value,
        vendor=f"- **Vendor:** {doc.vendor}" if doc.vendor is not None else "",
        description=_make_paragraph(doc.description),
        references=make_references(doc.references),
        support_info='success "Fully supported"'
        if doc.fully_supported
        else 'warning "Partially supported"',
        limitations=make_limitations(doc.limitations),
    )
